# encoding: utf-8
from __future__ import unicode_literals
from email.header import Header

import uuid

from os.path import basename

import requests
from mimetypes import guess_type
from email.mime.base import MIMEBase
from email.encoders import encode_base64
from emails.compat import urlparse
from emails.compat import string_types, to_bytes

# class FileNotFound(Exception):
#    pass

MIMETYPE_UNKNOWN = 'application/unknown'


def fix_content_type(content_type, t='image'):
    if not content_type:
        return "%s/unknown" % t
    else:
        return content_type


class BaseFile(object):

    """
    Store base "attachment-file" information.
    """

    def __init__(self, **kwargs):
        """
        uri and filename are connected properties.
        if no filename set, filename extracted from uri.
        if no uri, but filename set, then uri==filename
        """
        self.uri = kwargs.get('uri', None)
        self.absolute_url = kwargs.get('absolute_url', None) or self.uri
        self.filename = kwargs.get('filename', None)
        self.data = kwargs.get('data', None)
        self._mime_type = kwargs.get('mime_type', None)
        self._headers = kwargs.get('headers', None)
        self._content_disposition = kwargs.get('content_disposition', None)
        self.subtype = kwargs.get('subtype', None)
        self.local_loader = kwargs.get('local_loader', None)
        self.id = id

    def as_dict(self, fields=None):
        fields = fields or ('uri', 'absolute_url', 'filename', 'data',
                            'mime_type', 'content_disposition', 'subtype')
        return dict([(k, getattr(self, k)) for k in fields])

    def get_data(self):
        _data = getattr(self, '_data', None)
        if isinstance(_data, string_types):
            return _data
        elif hasattr(_data, 'read'):
            return _data.read()
        else:
            return _data

    def set_data(self, value):
        self._data = value

    data = property(get_data, set_data)

    def get_uri(self):
        _uri = getattr(self, '_uri', None)
        if _uri is None:
            _filename = getattr(self, '_filename', None)
            if _filename:
                _uri = self._uri = _filename
        return _uri

    def set_uri(self, value):
        self._uri = value

    uri = property(get_uri, set_uri)

    def get_filename(self):
        _filename = getattr(self, '_filename', None)
        if _filename is None:
            _uri = getattr(self, '_uri', None)
            if _uri:
                parsed_path = urlparse.urlparse(_uri)
                _filename = basename(parsed_path.path)
                if not _filename:
                    _filename = str(uuid.uuid4())
                self._filename = _filename
        return _filename

    def set_filename(self, value):
        self._filename = value

    filename = property(get_filename, set_filename)

    def get_mime_type(self):
        r = getattr(self, '_mime_type', None)
        if r is None:
            filename = self.filename
            if filename:
                r = self._mime_type = guess_type(filename)[0]
        if not r:
            r = MIMETYPE_UNKNOWN
        self._mime_type = r
        return r

    mime_type = property(get_mime_type)

    def get_content_disposition(self):
        return getattr(self, '_content_disposition', None)

    def set_content_disposition(self, value):
        self._content_disposition = value

    content_disposition = property(get_content_disposition, set_content_disposition)

    @property
    def mime(self):
        if self.content_disposition is None:
            return None
        _mime = getattr(self, '_cached_mime', None)
        if _mime is None:
            filename = str(Header(self.filename, 'utf-8'))
            self._cached_mime = _mime = MIMEBase(*self.mime_type.split('/', 1))
            _mime.set_payload(to_bytes(self.data))
            encode_base64(_mime)
            _mime.add_header('Content-Disposition',
                             self.content_disposition,
                             filename=filename)
            if self.content_disposition == 'inline':
                _mime.add_header('Content-ID', '<{0}>'.format(filename))
        return _mime

    def reset_mime(self):
        self._mime = None

    def fetch(self):
        pass


class LazyHTTPFile(BaseFile):

    def __init__(self, fetch_params=None, **kwargs):
        BaseFile.__init__(self, **kwargs)
        self.fetch_params = dict(allow_redirects=True, verify=False)
        if fetch_params:
            self.fetch_params.update(fetch_params)
        self._fetched = False

    def fetch(self):
        if (not self._fetched) and self.uri:
            if self.local_loader:
                data = self.local_loader[self.uri]

                if data:
                    self._fetched = True
                    self._data = data
                    return

            r = requests.get(self.absolute_url or self.uri, **self.fetch_params)
            if r.status_code == 200:
                self._data = r.content
                self._headers = r.headers
                self._mime_type = fix_content_type(r.headers.get('content-type'), t='unknown')
                self._fetched = True

    def get_data(self):
        self.fetch()
        return self._data or ''

    def set_data(self, v):
        self._data = v

    data = property(get_data, set_data)

    @property
    def mime_type(self):
        self.fetch()
        return self._mime_type

    @property
    def headers(self):
        self.fetch()
        return self._headers



# coding: utf-8
from __future__ import unicode_literals

import time
from functools import wraps

from dateutil.parser import parse as dateutil_parse
from email.header import Header
from email.utils import formatdate, getaddresses

from emails.compat import string_types, to_unicode, is_callable, to_bytes

from .utils import SafeMIMEText, SafeMIMEMultipart, sanitize_address, parse_name_and_email
from .smtp import ObjectFactory, SMTPBackend
from .store import MemoryFileStore, BaseFile
from .signers import DKIMSigner

from .utils import load_email_charsets

load_email_charsets()  # sic!

ROOT_PREAMBLE = 'This is a multi-part message in MIME format.\n'


class BadHeaderError(ValueError):
    pass

# Header names that contain structured address data (RFC #5322)
ADDRESS_HEADERS = set(['from', 'sender', 'reply-to', 'to', 'cc', 'bcc', 'resent-from', 'resent-sender', 'resent-to',
                   'resent-cc', 'resent-bcc'])


def renderable(f):
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        r = f(self, *args, **kwargs)
        render = getattr(r, 'render', None)
        if render:
            d = render(**(self.render_data or {}))
            return d
        else:
            return r

    return wrapper


class IncompleteMessage(Exception):
    pass


class Message(object):
    """
    Email class

        message = HtmlEmail()

    Message parts:
        * html
        * text
        * attachments

    """

    attachment_cls = BaseFile
    dkim_cls = DKIMSigner
    smtp_pool_factory = ObjectFactory
    smtp_cls = SMTPBackend
    filestore_cls = MemoryFileStore

    def __init__(self,
                 charset=None,
                 message_id=None,
                 date=None,
                 subject=None,
                 mail_from=None,
                 mail_to=None,
                 headers=None,
                 html=None,
                 text=None,
                 attachments=None):

        self._attachments = None
        self.charset = charset or 'utf-8'  # utf-8 is standard de-facto, yeah
        self._message_id = message_id
        self.set_subject(subject)
        self.set_date(date)
        self.set_mail_from(mail_from)
        self.set_mail_to(mail_to)
        self.set_headers(headers)
        self.set_html(html=html)  # , url=self.html_from_url)
        self.set_text(text=text)  # , url=self.text_from_url)
        self.render_data = {}
        self._dkim_signer = None
        if attachments:
            for a in attachments:
                self.attachments.add(a)

        self.after_build = None

    def set_mail_from(self, mail_from):
        # In: ('Alice', '<alice@me.com>' )
        self._mail_from = mail_from and parse_name_and_email(mail_from) or None

    def set_mail_to(self, mail_to):
        # Now we parse only one to-addr
        # TODO: parse list of to-addrs
        mail_to = mail_to and parse_name_and_email(mail_to)
        self._mail_to = mail_to and [mail_to, ] or []

    def get_mail_to(self):
        return self._mail_to

    mail_to = property(get_mail_to, set_mail_to)

    def set_headers(self, headers):
        self._headers = headers

    def set_html(self, html, url=None):
        if hasattr(html, 'read'):
            html = html.read()
        self._html = html
        self._html_url = url

    def set_text(self, text, url=None):
        if hasattr(text, 'read'):
            text = text.read()
        self._text = text
        self._text_url = url

    def attach(self, **kwargs):
        if 'content_disposition' not in kwargs:
            kwargs['content_disposition'] = 'attachment'
        self.attachments.add(kwargs)

    @classmethod
    def from_loader(cls, loader, template_cls=None, **kwargs):
        """
        Get html and attachments from HTTPLoader
        """
        message = cls(html=template_cls and template_cls(loader.html) or loader.html, **kwargs)
        for att in loader.filestore:
            message.attach(**att.as_dict())
        return message

    @property
    @renderable
    def html_body(self):
        return self._html

    @property
    @renderable
    def text_body(self):
        return self._text

    def set_subject(self, value):
        self._subject = value

    @renderable
    def get_subject(self):
        return self._subject

    subject = property(get_subject, set_subject)

    def render(self, **kwargs):
        self.render_data = kwargs

    @property
    def attachments(self):
        if self._attachments is None:
            self._attachments = self.filestore_cls(self.attachment_cls)
        return self._attachments

    def set_date(self, value):
        if isinstance(value, string_types):
            _d = dateutil_parse(value)
            value = time.mktime(_d.timetuple())
            value = formatdate(value, True)
        self._date = value

    def get_date(self):
        if self._date is False:
            return None
        timeval = self._date
        if timeval:
            if is_callable(timeval):
                timeval = timeval()
        elif timeval is None:
            timeval = formatdate(None, True)
        return timeval

    message_date = property(get_date, set_date)

    def message_id(self):
        mid = self._message_id
        if mid is False:
            return None
        return is_callable(mid) and mid() or mid

    def encode_header(self, value):
        value = to_unicode(value, charset=self.charset)
        if isinstance(value, string_types):
            value = value.rstrip()
            _r = Header(value, self.charset)
            return str(_r)
        else:
            return value

    def encode_name_header(self, realname, email):
        if realname:
            r = "%s <%s>" % (self.encode_header(realname), email)
            return r
        else:
            return email

    def set_header(self, msg, key, value, encode=True):

        if value is None:
            # TODO: may be remove header here ?
            return

        # Prevent header injection
        if '\n' in value or '\r' in value:
            raise BadHeaderError("Header values can't contain newlines (got %r for header %r)" % (value, key))

        if key.lower() in ADDRESS_HEADERS:
            value = ', '.join(sanitize_address(addr, self.charset)
                              for addr in getaddresses((value,)))

        msg[key] = encode and self.encode_header(value) or value

    def _build_message(self, message_cls=None):

        message_cls = message_cls or SafeMIMEMultipart
        msg = message_cls()

        msg.preamble = ROOT_PREAMBLE

        self.set_header(msg, 'Date', self.message_date, encode=False)
        self.set_header(msg, 'Message-ID', self.message_id(), encode=False)

        if self._headers:
            for (name, value) in self._headers.items():
                self.set_header(msg, name, value)

        subject = self.subject
        if subject is None:
            raise IncompleteMessage("Message must have 'subject'")
        self.set_header(msg, 'Subject', subject)

        mail_from = self.encode_name_header(*self._mail_from)
        # if mail_from is None:
        #    raise IncompleteMessage("Message must have 'mail_from'")
        self.set_header(msg, 'From', mail_from, encode=False)

        mail_to = self._mail_to and self.encode_name_header(*self._mail_to[0]) or None
        self.set_header(msg, 'To', mail_to, encode=False)

        msgalt = SafeMIMEMultipart('alternative')
        msg.attach(msgalt)

        _text = self.text_body
        _html = self.html_body

        if not (_html or _text):
            raise ValueError("Message must contain 'html' or 'text' part")

        if _text:
            msgtext = SafeMIMEText(_text, 'plain', charset=self.charset)
            msgtext.set_charset(self.charset)
            msgalt.attach(msgtext)

        if _html:
            msghtml = SafeMIMEText(_html, 'html', charset=self.charset)
            msghtml.set_charset(self.charset)
            msgalt.attach(msghtml)

        for f in self.attachments:
            msgfile = f.mime
            if msgfile:
                msg.attach(msgfile)

        if self.after_build:
            self.after_build(self, msg)

        return msg

    def message(self, message_cls=None):
        msg = self._build_message(message_cls=message_cls)
        if self._dkim_signer:
            msg_str = msg.as_string()
            dkim_header = self._dkim_signer.get_sign_header(to_bytes(msg_str))
            if dkim_header:
                msg._headers.insert(0, dkim_header)
        return msg

    def as_string(self):
        # self.as_string() is not equialent self.message().as_string()
        # self.as_string() gets one less message-to-string conversions for dkim
        msg = self._build_message()
        r = msg.as_string()
        if self._dkim_signer:
            dkim_header = self._dkim_signer.get_sign(to_bytes(r))
            if dkim_header:
                r = dkim_header + r
        return r

    @property
    def smtp_pool(self):
        pool = getattr(self, '_smtp_pool', None)
        if pool is None:
            pool = self._smtp_pool = self.smtp_pool_factory(cls=self.smtp_cls)
        return pool

    def dkim(self, **kwargs):
        self._dkim_signer = self.dkim_cls(**kwargs)

    def send(self,
             to=None,
             set_mail_to=True,
             mail_from=None,
             set_mail_from=False,
             render=None,
             smtp_mail_options=None,
             smtp_rcpt_options=None,
             smtp=None):

        if render is not None:
            self.render(**render)

        if smtp is None:
            smtp = {'host': 'localhost', 'port': 25, 'timeout': 5}

        if isinstance(smtp, dict):
            smtp = self.smtp_pool[smtp]

        if not hasattr(smtp, 'sendmail'):
            raise ValueError(
                "smtp must be a dict or an object with method 'sendmail'. got %s" % type(smtp))

        mail_to = to
        if mail_to:
            mail_to = parse_name_and_email(mail_to)
            to_addr = mail_to[1]
            if set_mail_to:
                self.set_mail_to(mail_to)

        else:
            to_addr = self._mail_to[0][1]

        if not to_addr:
            raise ValueError('No to-addr')

        if mail_from:
            if set_mail_from:
                self.set_mail_from(mail_from)
                from_addr = self._mail_from[1]
            else:
                mail_from = parse_name_and_email(mail_from)
                from_addr = mail_from[1]
        else:
            from_addr = self._mail_from[1]

        if not from_addr:
            raise ValueError('No from-addr')

        params = dict(from_addr=from_addr,
                      to_addrs=[to_addr, ],
                      msg=self)
        if smtp_mail_options:
            params['mail_options'] = smtp_mail_options

        if smtp_rcpt_options:
            params['rcpt_options'] = smtp_rcpt_options

        response = smtp.sendmail(**params)
        return response[0]


def html(**kwargs):
    return Message(**kwargs)




# encoding: utf-8
import os, os.path
import logging
from .htmlloader import HTTPLoader
from .fileloader import FileSystemLoader, ZipLoader
from .stylesheets import PageStylesheets


def from_url(url, **kwargs):
    loader = HTTPLoader()
    loader.load_url(url=url, **kwargs)
    return loader

load_url = from_url

def from_file(filename, **kwargs):
    return from_directory(directory=os.path.dirname(filename), index_file=os.path.basename(filename), **kwargs)

def from_directory(directory, index_file=None, **kwargs):
    loader = HTTPLoader()
    local_loader = FileSystemLoader(searchpath=directory)
    index_file_name = local_loader.find_index_file(index_file)
    dirname, basename = os.path.split(index_file_name)
    if dirname:
        local_loader.base_path = dirname
    loader.load_file(local_loader[basename], local_loader=local_loader,  **kwargs)
    return loader

def from_zip(zip_file, **kwargs):
    loader = HTTPLoader()
    local_store = ZipLoader(file=zip_file)
    index_file_name = local_store.find_index_file()
    dirname, basename = os.path.split(index_file_name)
    if dirname:
        local_store.base_path = dirname
    logging.debug('from_zip: found index file: %s', index_file_name)
    loader.load_file(local_store[basename], local_loader=local_store,  **kwargs)
    return loader

def from_string(html, css=None, **kwargs):
    loader = HTTPLoader()
    loader.load_string(html=html, css=css,  **kwargs)
    return loader




# encoding: utf-8
from __future__ import unicode_literals
import string


class BaseTemplate:

    def __init__(self, template_text, **kwargs):
        self.template = template_text

    def render(self, **kwargs):
        raise NotImplemented


class StringTemplate:
    """
    string.Template is very basic template engine.
    Do not use it really.
    """
    def __init__(self, template_text, safe_substitute=True, **kwargs):
        self.template = string.Template(template_text)
        self.safe_substitute = safe_substitute

    def render(self, **kwargs):
        r = self.safe_substitute and self.template.safe_substitute or self.template.substitute
        return r(**kwargs)


# -*- coding: utf-8 -*-
from __future__ import unicode_literals
"""
pythoncompat
"""

from .orderedset import OrderedSet

#from . import _urlparse as urlparse


import sys

# -------
# Pythons
# -------

# Syntax sugar.
_ver = sys.version_info

#: Python 2.x?
is_py2 = (_ver[0] == 2)

#: Python 3.x?
is_py3 = (_ver[0] == 3)

#: Python 3.0.x
is_py30 = (is_py3 and _ver[1] == 0)

#: Python 3.1.x
is_py31 = (is_py3 and _ver[1] == 1)

#: Python 3.2.x
is_py32 = (is_py3 and _ver[1] == 2)

#: Python 3.3.x
is_py33 = (is_py3 and _ver[1] == 3)

#: Python 3.4.x
is_py34 = (is_py3 and _ver[1] == 4)

#: Python 2.7.x
is_py27 = (is_py2 and _ver[1] == 7)

#: Python 2.6.x
is_py26 = (is_py2 and _ver[1] == 6)

#: Python 2.5.x
is_py25 = (is_py2 and _ver[1] == 5)

#: Python 2.4.x
is_py24 = (is_py2 and _ver[1] == 4)


# ---------
# Platforms
# ---------


# Syntax sugar.
_ver = sys.version.lower()

is_pypy = ('pypy' in _ver)
is_jython = ('jython' in _ver)
is_ironpython = ('iron' in _ver)

# Assume CPython, if nothing else.
is_cpython = not any((is_pypy, is_jython, is_ironpython))

# Windows-based system.
is_windows = 'win32' in str(sys.platform).lower()

# Standard Linux 2+ system.
is_linux = ('linux' in str(sys.platform).lower())
is_osx = ('darwin' in str(sys.platform).lower())
is_hpux = ('hpux' in str(sys.platform).lower())   # Complete guess.
is_solaris = ('solar==' in str(sys.platform).lower())   # Complete guess.

# ---------
# Specifics
# ---------

if is_py2:

    unichr = unichr
    text_type = unicode
    string_types = (str, unicode)
    integer_types = (int, long)
    int_to_byte = chr

    import urlparse
    from .ordereddict import OrderedDict

    from StringIO import StringIO
    from cStringIO import StringIO as BytesIO
    NativeStringIO = BytesIO

    def to_native(x, charset=sys.getdefaultencoding(), errors='strict'):
        if x is None or isinstance(x, str):
            return x
        return x.encode(charset, errors)

    def is_callable(x):
        return callable(x)

    def to_bytes(x, charset=sys.getdefaultencoding(), errors='strict'):
        if x is None:
            return None
        if isinstance(x, (bytes, bytearray, buffer)):
            return bytes(x)
        if isinstance(x, unicode):
            return x.encode(charset, errors)
        raise TypeError('Expected bytes')


elif is_py3:
    import urllib.parse as urlparse
    from collections import OrderedDict, Callable
    from io import StringIO, BytesIO
    NativeStringIO = StringIO

    unichr = chr
    text_type = str
    string_types = (str, )
    integer_types = (int, )

    def to_native(x, charset=sys.getdefaultencoding(), errors='strict'):
        if x is None or isinstance(x, str):
            return x
        return x.decode(charset, errors)

    def is_callable(x):
        return isinstance(x, Callable)

    def to_bytes(x, charset=sys.getdefaultencoding(), errors='strict'):
        if x is None:
            return None
        if isinstance(x, (bytes, bytearray, memoryview)):
            return bytes(x)
        if isinstance(x, str):
            return x.encode(charset, errors)
        raise TypeError('Expected bytes')


def to_unicode(x, charset=sys.getdefaultencoding(), errors='strict',
               allow_none_charset=False):
    if x is None:
        return None
    if not isinstance(x, bytes):
        return text_type(x)
    if charset is None and allow_none_charset:
        return x
    return x.decode(charset, errors)


# encoding: utf-8
from __future__ import unicode_literals
import posixpath
import os.path
import logging
import re
import warnings
from cssutils import CSSParser
from lxml import etree
from premailer import Premailer
from premailer.premailer import ExternalNotFoundError

import emails
from emails.compat import urlparse, to_unicode, to_bytes, text_type
from emails.store import MemoryFileStore, LazyHTTPFile
from .loader.local_store import FileNotFound


class LocalPremailer(Premailer):

    def __init__(self, html, local_loader=None, **kw):
        if 'preserve_internal_links' not in kw:
            kw['preserve_internal_links'] = True
        self.local_loader = local_loader
        super(LocalPremailer, self).__init__(html=html, **kw)

    def _load_external(self, url):
        """
        loads an external stylesheet from a remote url or local store
        """
        if url.startswith('//'):
            # then we have to rely on the base_url
            if self.base_url and 'https://' in self.base_url:
                url = 'https:' + url
            else:
                url = 'http:' + url

        if url.startswith('http://') or url.startswith('https://'):
            content = self._load_external_url(url)
        else:
            content = None

            if self.local_loader:
                try:
                    content = self.local_loader.get_source(url)
                except FileNotFound:
                    content = None

            if content is None:
                if self.base_url:
                    return self._load_external(urlparse.urljoin(self.base_url, url))
                else:
                    raise ExternalNotFoundError(url)

        return content


class HTMLParser(object):

    _cdata_regex = re.compile(r'\<\!\[CDATA\[(.*?)\]\]\>', re.DOTALL)

    def __init__(self, html, method="html"):
        self._html = html
        self._method = method
        self._tree = None

    @property
    def html(self):
        return self._html

    @property
    def tree(self):
        if self._tree is None:
            parser = self._method == 'xml' \
                         and etree.XMLParser(ns_clean=False, resolve_entities=False) \
                         or etree.HTMLParser()
            self._tree = etree.fromstring(self._html.strip(), parser)
        return self._tree

    def to_string(self, encoding='utf-8', **kwargs):
        out = etree.tostring(self.tree, encoding=encoding, method=self._method, **kwargs).decode(encoding)
        if self._method == 'xml':
            out = self._cdata_regex.sub(
                lambda m: '/*<![CDATA[*/%s/*]]>*/' % m.group(1),
                out
            )
        return out

    def apply_to_images(self, func, images=True, backgrounds=True, styles_uri=True):

        def _apply_to_style_uri(style_text, func):
            dirty = False
            parser = CSSParser().parseStyle(style_text)
            for prop in parser.getProperties(all=True):
                for value in prop.propertyValue:
                    if value.type == 'URI':
                        old_uri = value.uri
                        new_uri = func(old_uri, element=value)
                        if new_uri != old_uri:
                            dirty = True
                            value.uri = new_uri
            if dirty:
                return to_unicode(parser.cssText, 'utf-8')
            else:
                return style_text

        if images:
            # Apply to images from IMG tag
            for img in self.tree.xpath(".//img"):
                if 'src' in img.attrib:
                    img.attrib['src'] = func(img.attrib['src'], element=img)

        if backgrounds:
            # Apply to images from <tag background="X">
            for item in self.tree.xpath("//@background"):
                tag = item.getparent()
                tag.attrib['background'] = func(tag.attrib['background'], element=tag)

        if styles_uri:
            # Apply to style uri
            for item in self.tree.xpath("//@style"):
                tag = item.getparent()
                tag.attrib['style'] = _apply_to_style_uri(tag.attrib['style'], func=func)

    def apply_to_links(self, func):
        # Apply to images from IMG tag
        for a in self.tree.xpath(".//a"):
            if 'href' in a.attrib:
                a.attrib['href'] = func(a.attrib['href'], element=a)

    def add_content_type_meta(self, content_type="text/html", charset="utf-8", element_cls=etree.Element):

        def _get_content_type_meta(head):
            content_type_meta = None
            for meta in head.find('meta') or []:
                http_equiv = meta.get('http-equiv', None)
                if http_equiv and (http_equiv.lower() == 'content_type'):
                    content_type_meta = meta
                    break
            if content_type_meta is None:
                content_type_meta = element_cls('meta')
                head.append(content_type_meta)
            return content_type_meta

        head = self.tree.find('head')
        if head is None:
            logging.warning('HEAD not found. This should not happen. Skip.')
            return

        meta = _get_content_type_meta(head)
        meta.set('content', '%s; charset=%s' % (content_type, charset))
        meta.set('http-equiv', "Content-Type")


class BaseTransformer(HTMLParser):

    UNSAFE_TAGS = ['script', 'object', 'iframe', 'frame', 'base', 'meta', 'link', 'style']

    attachment_store_cls = MemoryFileStore
    attachment_file_cls = LazyHTTPFile

    def __init__(self, html, local_loader=None,
                 attachment_store=None,
                 requests_params=None, method="html", base_url=None):

        HTMLParser.__init__(self, html=html, method=method)

        self.attachment_store = attachment_store if attachment_store is not None else self.attachment_store_cls()
        self.local_loader = local_loader
        self.base_url = base_url
        self.requests_params = requests_params

    def get_absolute_url(self, url):

        if not self.base_url:
            return url

        if url.startswith('//'):
            if 'https://' in self.base_url:
                url = 'https:' + url
            else:
                url = 'http:' + url
            return url

        if not (url.startswith('http://') or url.startswith('https://')):
            url = urlparse.urljoin(self.base_url, posixpath.normpath(url))

        return url

    def _load_attachment_func(self, uri, element=None, **kw):
        #
        # Load uri from remote url or from local_store
        # Return local uri
        #
        attachment = self.attachment_store.by_uri(uri)
        if attachment is None:
            attachment = self.attachment_file_cls(
                uri=uri,
                absolute_url=self.get_absolute_url(uri),
                local_loader=self.local_loader,
                requests_args=self.requests_params)
            self.attachment_store.add(attachment)
        return attachment.filename

    def remove_unsafe_tags(self):
        for tag in self.UNSAFE_TAGS:
            for el in self.tree.xpath(".//%s" % tag):
                parent = el.getparent()
                if parent is not None:
                    parent.remove(el)

    def load_and_transform(self,
                           css_inline=True,
                           remove_unsafe_tags=True,
                           make_links_absolute=True,
                           set_content_type_meta=True,
                           update_stylesheet=True,
                           load_images=True,
                           images_inline=True,
                           **kw):

        if not make_links_absolute:
            # Now we use Premailer that always makes links absolute
            warnings.warn("make_links_absolute=False is deprecated.", DeprecationWarning)

        if not css_inline:
            # Premailer always makes inline css.
            warnings.warn("css_inline=False is deprecated.", DeprecationWarning)

        if update_stylesheet:
            # Premailer has no such feature.
            warnings.warn("update_stylesheet=True is deprecated.", DeprecationWarning)

        # 1. Premailer make some transformations on self.root tree:
        #  - load external css and make css inline
        #  - make absolute href and src if base_url is set
        premailer = LocalPremailer(html=self.tree,
                                   local_loader=self.local_loader,
                                   method=self._method,
                                   base_url=self.base_url,
                                   **kw)
        premailer.transform()

        # 2. Load linked images and transform links
        if load_images:
            self.apply_to_images(self._load_attachment_func)

        # 3. Remove unsafe tags is requested
        if remove_unsafe_tags:
            self.remove_unsafe_tags()

        # 4. Set <meta> content-type
        if set_content_type_meta:
            # TODO: may be remove this ?
            self.add_content_type_meta()

        # 5. Make images inline
        if load_images and images_inline:
            for a in self.attachment_store:
                a.is_inline = True
            self.synchronize_inline_images()

    def synchronize_inline_images(self, inline_names=None, non_inline_names=None):
        """
        Set img src in html for images, marked as "inline" in attachments_store
        """

        if inline_names is None or non_inline_names is None:

            inline_names = {}
            non_inline_names = {}

            for a in self.attachment_store:
                if a.is_inline:
                    inline_names[a.filename] = a.content_id
                else:
                    non_inline_names[a.content_id] = a.filename

        def _src_update_func(src, **kw):
            if src.startswith('cid:'):
                content_id = src[4:]
                if content_id in non_inline_names:
                    return non_inline_names[content_id]
            else:
                if src in inline_names:
                    return 'cid:'+inline_names[src]
            return src

        self.apply_to_images(_src_update_func)


class Transformer(BaseTransformer):

    @staticmethod
    def from_message(cls, message, **kw):
        return cls(html=message.html, attachment_store=message.attachments, **kw)

    def to_message(self, message=None):
        if message is None:
            message = emails.Message()
        message.html_body = self.to_string()
        # TODO: Copy attachments may be.
        message._attachments = self.attachment_store


class MessageTransformer(BaseTransformer):

    def __init__(self, message, **kw):
        self.message = message
        params = {'html': message._html, 'attachment_store': message.attachments}
        params.update(kw)
        BaseTransformer.__init__(self, **params)

    def save(self):
        self.message._html = self.to_string()


# coding: utf-8
from __future__ import unicode_literals
"""
python emails library
~~~~~~~~~~~~~~~~~~~~~

emails is a python library for dealing with html-emails.

Usage:

   >>> import emails
   >>> message = emails.html(html="<p>Hi!<br>Here is your receipt...",
                          subject="Your receipt No. 567098123",
                          mail_from=('Some Store', 'store@somestore.com'))
   >>> message.send( to = 's@lavr.me', smtp={ 'host': 'aspmx.l.google.com' } )


More examples is at <https://github.com/lavr/python-emails/README.rst>.

:copyright: (c) 2013 by Sergey Lavrinenko.
:license: Apache 2.0, see LICENSE for more details.

"""

__title__ = 'emails'
__version__ = '0.1.13'
__author__ = 'Sergey Lavrinenko'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2013 Sergey Lavrinenko'

from .message import Message, html
from .utils import MessageID





# encoding: utf-8
from __future__ import unicode_literals
import posixpath
import os.path
import logging

from lxml import etree
import requests

from emails.compat import urlparse, to_unicode, to_bytes, text_type
from emails.store import MemoryFileStore, LazyHTTPFile
from .stylesheets import PageStylesheets, StyledTagWrapper
from .cssinliner import CSSInliner
from .helpers import guess_charset
from .wrappers import TAG_WRAPPER, CSS_WRAPPER
from . import helpers


class HTTPLoaderError(Exception):
    pass


class HTTPLoader:
    """
    HTML loader loads single html page and store it as some sort of web archive:
        * loads html page
        * loads linked images
        * loads linked css and images from css
        * converts css to inline html styles
    """

    USER_AGENT = 'python-emails/1.0'

    UNSAFE_TAGS = set(['script', 'object', 'iframe', 'frame', 'base', 'meta', 'link', 'style'])
    TAGS_WITH_BACKGROUND = set(['td', 'tr', 'th', 'body'])
    TAGS_WITH_IMAGES = TAGS_WITH_BACKGROUND.union(set(['img', ]))
    CSS_MEDIA = ['', 'screen', 'all', 'email']

    tag_link_cls = {
        'a': TAG_WRAPPER('href'),
        'link': TAG_WRAPPER('href'),
        'img': TAG_WRAPPER('src'),
        'td': TAG_WRAPPER('background'),
        'table': TAG_WRAPPER('background'),
        'th': TAG_WRAPPER('background'),
    }

    css_link_cls = CSS_WRAPPER

    attached_image_cls = LazyHTTPFile
    filestore_cls = MemoryFileStore

    def __init__(self, filestore=None, encoding='utf-8', fetch_params=None):
        self.filestore = filestore or self.filestore_cls()
        self.encoding = encoding
        self.fetch_params = fetch_params
        self.stylesheets = PageStylesheets()
        self.base_url = None
        self._attachments = None
        self.local_loader = None

    def _fetch(self, url, valid_http_codes=(200, ), fetch_params=None):
        _params = dict(allow_redirects=True, verify=False,
                       headers={'User-Agent': self.USER_AGENT})
        fetch_params = fetch_params or self.fetch_params
        if fetch_params:
            _params.update(fetch_params)
        response = requests.get(url, **_params)
        if valid_http_codes and (response.status_code not in valid_http_codes):
            raise HTTPLoaderError('Error loading url: %s. HTTP status: %s' % (url, response.http_status))
        return response

    def get_html_tree(self):
        return self._html_tree

    def set_html_tree(self, value):
        self._html_tree = value
        self._html = None  # We never actually store html, only cached html_tree render

    html_tree = property(get_html_tree, set_html_tree)

    def tag_has_link(self, tag):
        return tag in self.tag_link_cls

    def start_load_url(self, url, base_url=None):
        """
        Set some params and load start page
        """

        # Load start page
        response = self._fetch(url, valid_http_codes=(200, ), fetch_params=self.fetch_params)
        self.start_url = url
        self.base_url = base_url or url  # Fixme: split base_url
        self.headers = response.headers
        content = response.content
        self.html_encoding = guess_charset(response.headers, content)
        if self.html_encoding:
            content = to_unicode(content, self.html_encoding)
        else:
            content = to_unicode(content)
        content = content.replace('\r\n', '\n')  # Remove \r, or we'll get &#13;
        self.html_content = content

    def start_load_file(self, html, encoding="utf-8"):
        """
        Set some params and load start page
        """
        if hasattr(html, 'read'):
            html = html.read()

        if not isinstance(html, text_type):
            html = to_unicode(html, encoding)

        html = html.replace('\r\n', '\n')  # Remove \r, or we'll get &#13;
        self.html_content = html
        self.html_encoding = encoding
        self.start_url = None
        self.base_url = None
        self.headers = None

    def start_load_string(self, html, css):
        self.html_content = html
        if css:
            self.stylesheets.append(text=css)
        self.html_encoding = 'utf-8'
        self.start_url = None
        self.base_url = None
        self.headers = None

    def make_html_tree(self):
        self.html_tree = etree.HTML(self.html_content, parser=etree.HTMLParser())
        # TODO: try another load methods, i.e. etree.fromstring(xml,
        # base_url="http://where.it/is/from.xml") ?

    def parse_html_tree(self, remove_unsafe_tags=True):

        # Parse html, load important tags

        self._a_links = []
        self._tags_with_links = []
        self._tags_with_images = []

        for el in self.html_tree.iter():

            if el.tag == 'img' or el.tag == 'a' or self.tag_has_link(el.tag):
                self.process_tag_with_link(el)

            if el.tag == 'base':
                self.base_url = el.get('href')  # TODO: can be relative link in BASE HREF ?

            elif el.tag == 'link':
                self.process_external_css_tag(el)

            elif el.tag == 'style':
                self.process_style_tag(el)

            # elif el.tag=='a':
            # self.process_a_tag( el )

            if el.get('style'):
                self.process_tag_with_style(el)

            if remove_unsafe_tags and (el.tag in self.UNSAFE_TAGS):
                # Remove unsafe tags
                # self._removed_unsafe.append(el)  # Save it for reports
                p = el.getparent()
                if p is not None:
                    p.remove(el)

        # now make concatenated stylesheet
        for prop in self.stylesheets.uri_properties:
            self.process_stylesheet_uri_property(prop)

        self.attach_all_images()

    def load_url(self, url, base_url=None, **kwargs):
        self.start_load_url(url=url, base_url=base_url)
        return self._load(**kwargs)

    def load_file(self, file, local_loader=None, **kwargs):
        self.local_loader = local_loader
        self.start_load_file(html=file)
        return self._load(**kwargs)

    def load_string(self, html, css, **kwargs):
        self.start_load_string(html=html, css=css)
        return self._load(**kwargs)

    def _load(self,
              css_inline=True,
              remove_unsafe_tags=True,
              make_links_absolute=False,
              set_content_type_meta=True,
              update_stylesheet=True,
              images_inline=False):

        self.make_html_tree()
        self.parse_html_tree(remove_unsafe_tags=remove_unsafe_tags)

        if make_links_absolute:
            [self.make_link_absolute(obj) for obj in self.iter_image_links()]
            [self.make_link_absolute(obj) for obj in self.iter_a_links()]

        if remove_unsafe_tags and update_stylesheet:
            self.stylesheets.attach_tag(self.insert_big_stylesheet())

        # self.process_attaches()

        # TODO: process images in self._tags_with_styles
        if css_inline:
            self.doinlinecss()

        if set_content_type_meta:
            self.set_content_type_meta()

        if images_inline:
            self.make_images_inline()

    def process_external_css_tag(self, el):
        """
        Process <link href="..." rel="stylesheet">
        """
        if el.get('rel', '') == 'stylesheet' and el.get('media', '') in self.CSS_MEDIA:
            url = el.get('href', '')
            if url:
                self.stylesheets.append(url=url,
                                        absolute_url=self.absolute_url(url),
                                        local_loader=self.local_loader)

    def process_style_tag(self, el):
        """
        Process: <style>...</style>
        """
        if el.text:
            self.stylesheets.append(text=el.text, url=self.start_url)

    def iter_image_links(self):
        return (_ for _ in self._tags_with_images)

    def iter_a_links(self):
        return (_ for _ in self._a_links)

    def process_tag_with_link(self, el):
        """
        Process IMG SRC, TABLE BACKGROUND, ...
        """
        obj = self.tag_link_cls[el.tag](el, encoding=self.html_encoding)
        if obj.link is None:
            return

        self._tags_with_links.append(obj)
        if el.tag in self.TAGS_WITH_IMAGES:
            lnk = obj.link
            if lnk is not None:
                self._tags_with_images.append(obj)
        elif el.tag == 'a':
            self._a_links.append(obj)

    def attach_all_images(self):
        for obj in self.iter_image_links():
            lnk = obj.link
            if lnk:
                self.attach_image(uri=lnk, absolute_url=self.absolute_url(lnk))

    def attach_image(self, uri, absolute_url, subtype=None):
        if uri not in self.filestore:
            self.filestore.add(self.attached_image_cls(
                uri=uri,
                absolute_url=absolute_url,
                local_loader=self.local_loader,
                subtype=subtype,
                fetch_params=self.fetch_params))

    def process_tag_with_style(self, el):
        t = StyledTagWrapper(el)
        for p in t.uri_properties():
            obj = self.css_link_cls(p, updateme=t)
            self._tags_with_links.append(obj)
            self._tags_with_images.append(obj)

    def process_stylesheet_uri_property(self, prop):
        obj = self.css_link_cls(prop)
        self._tags_with_links.append(obj)
        self._tags_with_images.append(obj)

    def make_link_absolute(self, obj):
        link = obj.link
        if link:
            obj.link = self.absolute_url(link)

    def make_images_inline(self):

        found_links = set()

        for img in self.iter_image_links():
            link = img.link
            found_links.add(link)
            file = self.filestore.by_uri(link, img.link_history)
            img.link = "cid:%s" % file.filename

        for file in self.filestore:
            if file.uri in found_links:
                file.content_disposition = 'inline'
            else:
                logging.debug('make_images_inline %s=none', file.uri)

    def set_content_type_meta(self):
        _tree = self.html_tree
        new_document = helpers.set_content_type_meta(_tree, element_cls=etree.Element)
        if _tree != new_document:
            # document may be updated here (i.e. html tag added)
            self.html_tree = new_document

    def insert_big_stylesheet(self):
        return helpers.add_body_stylesheet(self.html_tree, element_cls=etree.Element,
                                           tag="body", cssText="")

    def absolute_url(self, url, base_url=None):

        # In: some url
        # Out: (absolute_url, relative_url) based on self._base_url

        if base_url is None:
            base_url = self.base_url

        if base_url is None:
            return url

        parsed_url = urlparse.urlsplit(url)
        if parsed_url.scheme:
            # is absolute_url
            return url
        else:
            # http://xxx.com/../../style.css -> http://xxx.com/style.css
            # см. http://teethgrinder.co.uk/perm.php?a=Normalize-URL-path-python
            joined = urlparse.urljoin(self.base_url, url)
            url = urlparse.urlparse(joined)
            path = posixpath.normpath(url[2])
            return urlparse.urlunparse((url.scheme, url.netloc, path, url.params, url.query, url.fragment))

    def doinlinecss(self):
        self.html_tree = CSSInliner(css=self.stylesheets.stylesheet).transform(html=self.html_tree)

    @property
    def html(self):
        self.stylesheets.update_tag()
        self._html = etree.tostring(self.html_tree, encoding=self.encoding, method='xml')
        return to_unicode(self._html, self.encoding)

    @property
    def attachments_dict(self):
        return list(self.filestore.as_dict())

    def save_to_file(self, filename):
        #
        # Not very good example of link walking and file rename
        #

        path = os.path.abspath(filename)
        # Save images locally and replace all links to images in html
        files_dir = '_files'
        _rename_map = {}

        for obj in self.iter_image_links():
            uri = obj.link
            if uri is None:
                continue
            _new_uri = _rename_map.get(uri, None)
            if _new_uri is None:
                file = self.filestore.by_uri(uri, synonims=obj.link_history)
                if file is None:
                    logging.warning(
                        'file "%s" not found in attachments, this should not happen. skipping', uri)
                    continue
                _new_uri = _rename_map[uri] = os.path.join(files_dir, file.filename)
            obj.link = _new_uri

        try:
            os.makedirs(files_dir)
        except OSError:
            pass
        for attach in self.filestore:
            attach.fetch()
            new_uri = _rename_map.get(attach.uri)
            if new_uri:
                attach.uri = new_uri
                open(new_uri, 'wb').write(attach.data)

        f = open(filename, 'wb')
        f.write(to_bytes(self.html, 'utf-8'))
        f.close()




# encoding: utf-8
from __future__ import unicode_literals

__all__ = [ 'SMTPSender' ]

import smtplib
import logging
import threading
from functools import wraps

from .client import SMTPResponse, SMTPClientWithResponse, SMTPClientWithResponse_SSL
from emails.compat import to_bytes

logger = logging.getLogger(__name__)

class SMTPBackend:

    """
    SMTPSender is a wrapper for smtplib.SMTP class.
    Differences are:
    a) it transparently uses SSL or no-SSL connection
    b) sendmail method sends only one message, but returns more information
       about server response (i.e. response code)
    """

    DEFAULT_SOCKET_TIMEOUT = 5

    connection_cls = SMTPClientWithResponse
    connection_ssl_cls = SMTPClientWithResponse_SSL
    response_cls = SMTPResponse


    def __init__(self,
                 user=None,
                 password=None,
                 ssl=False,
                 tls=False,
                 debug=False,
                 fail_silently=True,
                 **kwargs):

        self.smtp_cls = ssl and self.connection_ssl_cls or self.connection_cls
        self.debug = debug
        self.ssl = ssl
        self.tls = tls
        if self.ssl and self.tls:
            raise ValueError(
                "ssl/tls are mutually exclusive, so only set "
                "one of those settings to True.")

        self.user = user
        self.password = password
        if 'timeout' not in kwargs:
            kwargs['timeout'] = self.DEFAULT_SOCKET_TIMEOUT
        self.smtp_cls_kwargs = kwargs

        self.host = kwargs.get('host')
        self.port = kwargs.get('port')
        self.fail_silently = fail_silently
        self.connection = None
        #self.local_hostname=DNS_NAME.get_fqdn()
        self._lock = threading.RLock()

    def open(self):
        #logger.debug('SMTPSender _connect')
        if self.connection is None:
            self.connection = self.smtp_cls(parent=self, **self.smtp_cls_kwargs)
            self.connection.initialize()
        return self.connection

    def close(self):
        """Closes the connection to the email server."""

        if self.connection is None:
            return

        try:
            self.connection.close()
        except:
                if self.fail_silently:
                    return
                raise
        finally:
            self.connection = None


    def make_response(self, exception=None):
        return self.response_cls(host=self.host, port=self.port, exception=exception)

    def retry_on_disconnect(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except smtplib.SMTPServerDisconnected as e:
                # If server disconected, just connect again
                logging.debug('SMTPServerDisconnected, retry once')
                self.close()
                self.open()
                return func(*args, **kwargs)
        return wrapper

    def _sendmail_on_connection(self, *args, **kwargs):
        return self.connection.sendmail(*args, **kwargs)

    def sendmail(self, from_addr, to_addrs, msg, mail_options=[], rcpt_options=[]):

        if not to_addrs: return False

        if not isinstance(to_addrs, (list, tuple)):
            to_addrs = [to_addrs, ]

        #from_addr = sanitize_address(from_addr, email_message.encoding)
        #to_addrs = [sanitize_address(addr, email_message.encoding) for addr in to_addrs]
        #message = email_message.message()
        #charset = message.get_charset().get_output_charset() if message.get_charset() else 'utf-8'

        try:
            self.open()
        except (IOError, smtplib.SMTPException) as e:
            logger.exception("Error connecting smtp server")
            response = self.make_response(exception = e)
            if not self.fail_silently:
                response.raise_if_needed()
            return [response, ]

        _sendmail = self.retry_on_disconnect(self._sendmail_on_connection)

        response = _sendmail(from_addr=from_addr,
                             to_addrs=to_addrs,
                             msg=to_bytes(msg.as_string(), 'utf-8'),
                             mail_options=mail_options,
                             rcpt_options=rcpt_options)

        if not self.fail_silently:
            [ r.raise_if_needed() for r in response ]

        return response




# encoding: utf-8
from __future__ import unicode_literals, print_function
import logging

from cssutils.css import CSSStyleSheet
from cssutils import CSSParser
import cssutils

from emails.compat import to_unicode


class PageStylesheets:

    """
    Store all html page styles and generates concatenated stylesheet
    """

    def __init__(self):
        self.urls = set()
        self._uri_properties = []
        self.sheets = []
        self.dirty = True
        self.element = None

    def update_tag(self):
        if self.element is not None:
            self._concatenate_sheets()
            cssText = self._cached_stylesheet.cssText
            cssText = cssText and to_unicode(cssText, 'utf-8') or ''
            self.element.text = cssText

    def attach_tag(self, element):
        self.element = element

    def append(self, url=None, text=None, absolute_url=None, local_loader=None):
        if (url is not None) and (url in self.urls):
            logging.debug('stylesheet url duplicate: %s', url)
            return
        self.sheets.append({'url': url, 'text': text, 'absolute_url': absolute_url or url,
                            'local_loader': local_loader})
        self.dirty = True

    def _concatenate_sheets(self):
        if self.dirty or (self._cached_stylesheet is None):
            r = CSSStyleSheet()
            uri_properties = []

            for d in self.sheets:
                local_loader = d.get('local_loader', None)
                text = d.get('text', None)
                uri = d.get('uri', None)
                absolute_url = d.get('absolute_url', None)

                if (text is None) and local_loader and uri:
                    text = local_loader[uri]

                if text:
                    sheet = CSSParser().parseString(text, href=absolute_url)
                else:
                    sheet = cssutils.parseUrl(href=absolute_url)

                for rule in sheet:
                    r.add(rule)
                    for p in _get_rule_uri_properties(rule):
                        uri_properties.append(p)

            self._uri_properties = uri_properties
            self._cached_stylesheet = r
            self.dirty = False

    @property
    def stylesheet(self):
        self._concatenate_sheets()
        return self._cached_stylesheet

    @property
    def uri_properties(self):
        self._concatenate_sheets()
        return self._uri_properties


class StyledTagWrapper:

    def __init__(self, el):
        self.el = el
        self.style = CSSParser().parseStyle(el.get('style'))

    def update(self):
        cssText = self.style.cssText
        if isinstance(cssText, str):
            cssText = to_unicode(cssText, 'utf-8')
        self.el.set('style', cssText)

    def uri_properties(self):
        for p in self.style.getProperties(all=True):
            for v in p.propertyValue:
                if v.type == 'URI':
                    yield v


# Stuff for extracting 'uri-properties' from CSS
# Inspired by cssutils examples

def _style_declarations(base):
    """recursive generator to find all CSSStyleDeclarations"""
    if hasattr(base, 'cssRules'):
        for rule in base.cssRules:
            for s in _style_declarations(rule):
                yield s
    elif hasattr(base, 'style'):
        yield base.style


def _get_rule_uri_properties(rule):
    for style in _style_declarations(rule):
        for p in style.getProperties(all=True):
            for v in p.propertyValue:
                if v.type == 'URI':
                    yield v


def get_stylesheets_uri_properties(sheet):
    for rule in sheet:
        for p in _get_rule_uri_properties(rule):
            yield p



# encoding: utf-8

# tag-with-link wrapper
from __future__ import unicode_literals
import logging
from emails.compat import OrderedSet, to_unicode


class ElementWithLink(object):

    LINK_ATTR_NAME = None

    def __init__(self, el, encoding=None):
        self.el = el
        self._link_history = OrderedSet()
        self.encoding = encoding

    def get_link(self):
        r = self.el.get(self.LINK_ATTR_NAME)
        if self.encoding:
            r = to_unicode(r, self.encoding)
        return r

    def set_link(self, new):
        _old = self.get_link()
        if _old != new:
            logging.debug('Update link %s => %s ', _old, new)
            self.el.set(self.LINK_ATTR_NAME, new)
            self._link_history.add(_old)

    link = property(get_link, set_link)

    @classmethod
    def make(cls, attr):
        def wrapper(el, encoding):
            r = cls(el, encoding=encoding)
            r.LINK_ATTR_NAME = attr
            return r
        return wrapper

    @property
    def link_history(self):
        return self._link_history


class A_link(ElementWithLink):
    # el is lxml.Element
    LINK_ATTR_NAME = 'href'


class Link_link(ElementWithLink):
    # el is lxml.Element
    LINK_ATTR_NAME = 'href'


class IMG_link(ElementWithLink):
    # el is lxml.Element
    LINK_ATTR_NAME = 'src'


class Background_link(ElementWithLink):
    LINK_ATTR_NAME = 'background'


class CSS_link(ElementWithLink):

    # el is cssutils style property

    def __init__(self, el, updateme=None, encoding=None):
        ElementWithLink.__init__(self, el)
        self.updateme = updateme
        self.encoding = encoding

    def get_link(self):
        r = self.el.uri
        if self.encoding:
            r = to_unicode(self.el.uri, self.encoding)
        return r

    def set_link(self, new):
        _old = self.el.uri
        if _old != new:
            logging.debug('Update link %s => %s ', _old, new)
            self.el.uri = new
            self._link_history.add(_old)
            if self.updateme:
                self.updateme.update()

    link = property(get_link, set_link)


def TAG_WRAPPER(attr):
    return ElementWithLink.make(attr)


CSS_WRAPPER = CSS_link



#!/usr/bin/python
# coding: utf-8

"""

Simple utility that imports html from url ang print generated rfc822 message to console.

Example usage:

    $ python make_rfc822.py --url=http://lavr.github.io/python-emails/tests/campaignmonitor-samples/sample-template/template-widgets.html
            --subject="Some subject"
            --from-name="Sergey Lavrinenko"
            --from-email=s@lavr.me
            --message-id-domain=localhost
            --send-test-email-to=sergei-nko@mail.ru
            --smtp-host=mxs.mail.ru
            --smtp-port=25

Copyright 2013  Sergey Lavrinenko <s@lavr.me>

"""

import sys
import logging
import json

import argparse

import emails
import emails.loader
from emails.template import JinjaTemplate as T


class MakeRFC822:

    def __init__(self, options):
        self.options = options

    def _headers_from_command_line(self):
        """
        --add-header "X-Source: AAA"
        """
        r = {}
        for s in self.options.add_headers:
            (k, v) = s.split(':', 1)
            r[k] = v
        return r

    def _get_message(self):

        options = self.options

        if options.message_id_domain:
            message_id = emails.utils.MessageID(domain=options.message_id_domain)
        else:
            message_id = None

        loader = emails.loader.from_url(url=options.url, images_inline=options.inline_images)


        message = emails.Message.from_loader(loader=loader,
                              headers= self._headers_from_command_line(), #{'X-Imported-From-URL': options.url },
                              template_cls=T,
                              mail_from=(options.from_name, options.from_email),
                              subject=T(unicode(options.subject, 'utf-8')),
                              message_id=message_id
                            )
        return message


    def _send_test_email(self, message):

        options = self.options

        if options.send_test_email_to:
            logging.debug("options.send_test_email_to YES")

            smtp_params = {}
            for k in ('host', 'port', 'ssl', 'user', 'password', 'debug'):
                smtp_params[k] = getattr(options, 'smtp_%s' % k, None)

            for mail_to in options.send_test_email_to.split(','):
                r = message.send(to=mail_to, smtp=smtp_params)
                logging.debug("mail_to=%s result=%s error=%s", mail_to, r, r.error)
                if r.error:
                    raise r.error

    def _start_batch(self):

        fn = self.options.batch
        if not fn: return None

        if fn=='-':
            f = sys.stdin
        else:
            f = open(fn, 'rb')

        def wrapper():
            for l in f.readlines():
                l = l.strip()
                if not l: continue
                # Magic is here
                try:
                    # Try to parse line as json
                    yield json.loads(l)
                except ValueError:
                    # If it is not json, we expect one word with '@' sign
                    assert len(l.split())==1
                    print l
                    login, domain = l.split('@') # ensure there is something email-like
                    yield {'to': l}

        return wrapper()

    def _generate_batch(self, batch, message):
        n = 0
        for values in batch:
            message.set_mail_to( values['to'] )
            message.render(**values.get('data', {}))
            s = message.as_string()
            n += 1
            logging.debug('Render email to %s', '%s.eml' % n)
            open('%s.eml' % n, 'wb').write(s)

    def main(self):

        options = self.options

        message = self._get_message()

        self._send_test_email(message)

        if self.options.batch:
            batch = self._start_batch()
            self._generate_batch(batch, message)
        else:
            batch = None
            if self.options.output_format=='eml':
                print(message.as_string())
            elif self.options.output_format=='html':
                print(message.html_body)








if __name__=="__main__":


    parser = argparse.ArgumentParser(description='Simple utility that imports html from url ang print generated rfc822 message to console.')

    parser.add_argument("-u", "--url", metavar="URL", dest="url", action="store", default=None, required=True)

    parser.add_argument("-f", "--from-email", metavar="EMAIL", dest="from_email", default=None, required=True)
    parser.add_argument("-n", "--from-name", metavar="NAME", dest="from_name", default=None, required=True)
    parser.add_argument("-s", "--subject", metavar="SUBJECT", dest="subject", default=None, required=True)
    parser.add_argument("--message-id-domain", dest="message_id_domain", default=None, required=True)

    parser.add_argument("--add-header", dest="add_headers", action='append', default=None, required=False)

    parser.add_argument("--inline-images", action="store_true", dest="inline_images", default=False)

    parser.add_argument("--output-format", dest="output_format", default='eml', choices=['eml', ])
    parser.add_argument("--log-level", dest="log_level", default="debug")

    parser.add_argument("--send-test-email-to", dest="send_test_email_to", default=None)
    parser.add_argument("--smtp-host", dest="smtp_host", default="localhost")
    parser.add_argument("--smtp-port", dest="smtp_port", default="25")
    parser.add_argument("--smtp-ssl", dest="smtp_ssl", action="store_true")
    parser.add_argument("--smtp-user", dest="smtp_user", default=None)
    parser.add_argument("--smtp-password", dest="smtp_password", default=None)
    parser.add_argument("--smtp-debug", dest="smtp_debug", action="store_true")

    parser.add_argument("--batch",       dest="batch",       default=None)
    parser.add_argument("--batch-start", dest="batch_start", default=None)
    parser.add_argument("--batch-limit", dest="batch_limit", default=None)

    options = parser.parse_args()

    logging.basicConfig( level=logging.getLevelName(options.log_level.upper()) )

    MakeRFC822(options=options).main()



# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the author be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
#
# Copyright (c) 2008 Greg Hewgill http://hewgill.com
#
# This has been modified from the original software.
# Copyright (c) 2011 William Grant <me@williamgrant.id.au>

import base64
import hashlib
import logging
import re
import time

from .canonicalization import (
    CanonicalizationPolicy,
    InvalidCanonicalizationPolicyError,
    )
from .crypto import (
    DigestTooLargeError,
    HASH_ALGORITHMS,
    parse_pem_private_key,
    parse_public_key,
    RSASSA_PKCS1_v1_5_sign,
    RSASSA_PKCS1_v1_5_verify,
    UnparsableKeyError,
    )
try:
  from .dnsplug import get_txt
except:
  def get_txt(s):
    raise RuntimeError("DKIM.verify requires DNS or dnspython module")
from .util import (
    get_default_logger,
    InvalidTagValueList,
    parse_tag_value,
    )

__all__ = [
    "DKIMException",
    "InternalError",
    "KeyFormatError",
    "MessageFormatError",
    "ParameterError",
    "Relaxed",
    "Simple",
    "DKIM",
    "sign",
    "verify",
]

Relaxed = b'relaxed'    # for clients passing dkim.Relaxed
Simple = b'simple'      # for clients passing dkim.Simple

def bitsize(x):
    """Return size of long in bits."""
    return len(bin(x)) - 2

class DKIMException(Exception):
    """Base class for DKIM errors."""
    pass

class InternalError(DKIMException):
    """Internal error in dkim module. Should never happen."""
    pass

class KeyFormatError(DKIMException):
    """Key format error while parsing an RSA public or private key."""
    pass

class MessageFormatError(DKIMException):
    """RFC822 message format error."""
    pass

class ParameterError(DKIMException):
    """Input parameter error."""
    pass

class ValidationError(DKIMException):
    """Validation error."""
    pass

def select_headers(headers, include_headers):
    """Select message header fields to be signed/verified.

    >>> h = [('from','biz'),('foo','bar'),('from','baz'),('subject','boring')]
    >>> i = ['from','subject','to','from']
    >>> select_headers(h,i)
    [('from', 'baz'), ('subject', 'boring'), ('from', 'biz')]
    >>> h = [('From','biz'),('Foo','bar'),('Subject','Boring')]
    >>> i = ['from','subject','to','from']
    >>> select_headers(h,i)
    [('From', 'biz'), ('Subject', 'Boring')]
    """
    sign_headers = []
    lastindex = {}
    for h in include_headers:
        assert h == h.lower()
        i = lastindex.get(h, len(headers))
        while i > 0:
            i -= 1
            if h == headers[i][0].lower():
                sign_headers.append(headers[i])
                break
        lastindex[h] = i
    return sign_headers

FWS = br'(?:\r?\n\s+)?'
RE_BTAG = re.compile(br'([;\s]b'+FWS+br'=)(?:'+FWS+br'[a-zA-Z0-9+/=])*(?:\r?\n\Z)?')

def hash_headers(hasher, canonicalize_headers, headers, include_headers,
                 sigheader, sig):
    """Update hash for signed message header fields."""
    sign_headers = select_headers(headers,include_headers)
    # The call to _remove() assumes that the signature b= only appears
    # once in the signature header
    cheaders = canonicalize_headers.canonicalize_headers(
        [(sigheader[0], RE_BTAG.sub(b'\\1',sigheader[1]))])
    # the dkim sig is hashed with no trailing crlf, even if the
    # canonicalization algorithm would add one.
    for x,y in sign_headers + [(x, y.rstrip()) for x,y in cheaders]:
        hasher.update(x)
        hasher.update(b":")
        hasher.update(y)
    return sign_headers

def validate_signature_fields(sig):
    """Validate DKIM-Signature fields.

    Basic checks for presence and correct formatting of mandatory fields.
    Raises a ValidationError if checks fail, otherwise returns None.

    @param sig: A dict mapping field keys to values.
    """
    mandatory_fields = (b'v', b'a', b'b', b'bh', b'd', b'h', b's')
    for field in mandatory_fields:
        if field not in sig:
            raise ValidationError("signature missing %s=" % field)

    if sig[b'v'] != b"1":
        raise ValidationError("v= value is not 1 (%s)" % sig[b'v'])
    if re.match(br"[\s0-9A-Za-z+/]+=*$", sig[b'b']) is None:
        raise ValidationError("b= value is not valid base64 (%s)" % sig[b'b'])
    if re.match(br"[\s0-9A-Za-z+/]+=*$", sig[b'bh']) is None:
        raise ValidationError(
            "bh= value is not valid base64 (%s)" % sig[b'bh'])
    # Nasty hack to support both str and bytes... check for both the
    # character and integer values.
    if b'i' in sig and (
        not sig[b'i'].endswith(sig[b'd']) or
        sig[b'i'][-len(sig[b'd'])-1] not in ('@', '.', 64, 46)):
        raise ValidationError(
            "i= domain is not a subdomain of d= (i=%s d=%s)" %
            (sig[b'i'], sig[b'd']))
    if b'l' in sig and re.match(br"\d{,76}$", sig[b'l']) is None:
        raise ValidationError(
            "l= value is not a decimal integer (%s)" % sig[b'l'])
    if b'q' in sig and sig[b'q'] != b"dns/txt":
        raise ValidationError("q= value is not dns/txt (%s)" % sig[b'q'])
    if b't' in sig and re.match(br"\d+$", sig[b't']) is None:
        raise ValidationError(
            "t= value is not a decimal integer (%s)" % sig[b't'])
    if b'x' in sig:
        if re.match(br"\d+$", sig[b'x']) is None:
            raise ValidationError(
                "x= value is not a decimal integer (%s)" % sig[b'x'])
        if b't' in sig and int(sig[b'x']) < int(sig[b't']):
            raise ValidationError(
                "x= value is less than t= value (x=%s t=%s)" %
                (sig[b'x'], sig[b't']))


def rfc822_parse(message):
    """Parse a message in RFC822 format.

    @param message: The message in RFC822 format. Either CRLF or LF is an accepted line separator.
    @return: Returns a tuple of (headers, body) where headers is a list of (name, value) pairs.
    The body is a CRLF-separated string.
    """
    headers = []
    lines = re.split(b"\r?\n", message)
    i = 0
    while i < len(lines):
        if len(lines[i]) == 0:
            # End of headers, return what we have plus the body, excluding the blank line.
            i += 1
            break
        if lines[i][0] in ("\x09", "\x20", 0x09, 0x20):
            headers[-1][1] += lines[i]+b"\r\n"
        else:
            m = re.match(br"([\x21-\x7e]+?):", lines[i])
            if m is not None:
                headers.append([m.group(1), lines[i][m.end(0):]+b"\r\n"])
            elif lines[i].startswith(b"From "):
                pass
            else:
                raise MessageFormatError("Unexpected characters in RFC822 header: %s" % lines[i])
        i += 1
    return (headers, b"\r\n".join(lines[i:]))



def fold(header):
    """Fold a header line into multiple crlf-separated lines at column 72.

    >>> fold(b'foo')
    'foo'
    >>> fold(b'foo  '+b'foo'*24).splitlines()[0]
    'foo  '
    >>> fold(b'foo'*25).splitlines()[-1]
    ' foo'
    >>> len(fold(b'foo'*25).splitlines()[0])
    72
    """
    i = header.rfind(b"\r\n ")
    if i == -1:
        pre = b""
    else:
        i += 3
        pre = header[:i]
        header = header[i:]
    while len(header) > 72:
        i = header[:72].rfind(b" ")
        if i == -1:
            j = 72
        else:
            j = i + 1
        pre += header[:j] + b"\r\n "
        header = header[j:]
    return pre + header

#: Hold messages and options during DKIM signing and verification.
class DKIM(object):
  # NOTE - the first 2 indentation levels are 2 instead of 4 
  # to minimize changed lines from the function only version.

  #: The U{RFC5322<http://tools.ietf.org/html/rfc5322#section-3.6>}
  #: complete list of singleton headers (which should
  #: appear at most once).  This can be used for a "paranoid" or
  #: "strict" signing mode.
  #: Bcc in this list is in the SHOULD NOT sign list, the rest could
  #: be in the default FROZEN list, but that could also make signatures 
  #: more fragile than necessary.  
  #: @since: 0.5
  RFC5322_SINGLETON = ('date','from','sender','reply-to','to','cc','bcc',
        'message-id','in-reply-to','references')

  #: Header fields to protect from additions by default.
  #: 
  #: The short list below is the result more of instinct than logic.
  #: @since: 0.5
  FROZEN = ('from','date','subject')

  #: The rfc4871 recommended header fields to sign
  #: @since: 0.5
  SHOULD = (
    'sender', 'reply-to', 'subject', 'date', 'message-id', 'to', 'cc',
    'mime-version', 'content-type', 'content-transfer-encoding', 'content-id',
    'content- description', 'resent-date', 'resent-from', 'resent-sender',
    'resent-to', 'resent-cc', 'resent-message-id', 'in-reply-to', 'references',
    'list-id', 'list-help', 'list-unsubscribe', 'list-subscribe', 'list-post',
    'list-owner', 'list-archive'
  )

  #: The rfc4871 recommended header fields not to sign.
  #: @since: 0.5
  SHOULD_NOT = (
    'return-path', 'received', 'comments', 'keywords', 'bcc', 'resent-bcc',
    'dkim-signature'
  )

  #: Create a DKIM instance to sign and verify rfc5322 messages.
  #:
  #: @param message: an RFC822 formatted message to be signed or verified
  #: (with either \\n or \\r\\n line endings)
  #: @param logger: a logger to which debug info will be written (default None)
  #: @param signature_algorithm: the signing algorithm to use when signing
  def __init__(self,message=None,logger=None,signature_algorithm=b'rsa-sha256',
        minkey=1024):
    self.set_message(message)
    if logger is None:
        logger = get_default_logger()
    self.logger = logger
    if signature_algorithm not in HASH_ALGORITHMS:
        raise ParameterError(
            "Unsupported signature algorithm: "+signature_algorithm)
    self.signature_algorithm = signature_algorithm
    #: Header fields which should be signed.  Default from RFC4871
    self.should_sign = set(DKIM.SHOULD)
    #: Header fields which should not be signed.  The default is from RFC4871.
    #: Attempting to sign these headers results in an exception.
    #: If it is necessary to sign one of these, it must be removed
    #: from this list first.
    self.should_not_sign = set(DKIM.SHOULD_NOT)
    #: Header fields to sign an extra time to prevent additions.
    self.frozen_sign = set(DKIM.FROZEN)
    #: Minimum public key size.  Shorter keys raise KeyFormatError. The
    #: default is 1024
    self.minkey = minkey

  def add_frozen(self,s):
    """ Add headers not in should_not_sign to frozen_sign.
    @param s: list of headers to add to frozen_sign
    @since: 0.5

    >>> dkim = DKIM()
    >>> dkim.add_frozen(DKIM.RFC5322_SINGLETON)
    >>> sorted(dkim.frozen_sign)
    ['cc', 'date', 'from', 'in-reply-to', 'message-id', 'references', 'reply-to', 'sender', 'subject', 'to']
    """
    self.frozen_sign.update(x.lower() for x in s
        if x.lower() not in self.should_not_sign)

  #: Load a new message to be signed or verified.
  #: @param message: an RFC822 formatted message to be signed or verified
  #: (with either \\n or \\r\\n line endings)
  #: @since: 0.5
  def set_message(self,message):
    if message:
      self.headers, self.body = rfc822_parse(message)
    else:
      self.headers, self.body = [],''
    #: The DKIM signing domain last signed or verified.
    self.domain = None
    #: The DKIM key selector last signed or verified.
    self.selector = 'default'
    #: Signature parameters of last sign or verify.  To parse
    #: a DKIM-Signature header field that you have in hand,
    #: use L{dkim.util.parse_tag_value}.
    self.signature_fields = {}
    #: The list of headers last signed or verified.  Each header
    #: is a name,value tuple.  FIXME: The headers are canonicalized.
    #: This could be more useful as original headers.
    self.signed_headers = []
    #: The public key size last verified.
    self.keysize = 0

  def default_sign_headers(self):
    """Return the default list of headers to sign: those in should_sign or
    frozen_sign, with those in frozen_sign signed an extra time to prevent
    additions.
    @since: 0.5"""
    hset = self.should_sign | self.frozen_sign
    include_headers = [ x for x,y in self.headers
        if x.lower() in hset ]
    return include_headers + [ x for x in include_headers
        if x.lower() in self.frozen_sign]

  def all_sign_headers(self):
    """Return header list of all existing headers not in should_not_sign.
    @since: 0.5"""
    return [x for x,y in self.headers if x.lower() not in self.should_not_sign]

  #: Sign an RFC822 message and return the DKIM-Signature header line.
  #:
  #: The include_headers option gives full control over which header fields
  #: are signed.  Note that signing a header field that doesn't exist prevents
  #: that field from being added without breaking the signature.  Repeated
  #: fields (such as Received) can be signed multiple times.  Instances
  #: of the field are signed from bottom to top.  Signing a header field more
  #: times than are currently present prevents additional instances
  #: from being added without breaking the signature.
  #:
  #: The length option allows the message body to be appended to by MTAs
  #: enroute (e.g. mailing lists that append unsubscribe information)
  #: without breaking the signature.
  #:
  #: The default include_headers for this method differs from the backward
  #: compatible sign function, which signs all headers not 
  #: in should_not_sign.  The default list for this method can be modified 
  #: by tweaking should_sign and frozen_sign (or even should_not_sign).
  #: It is only necessary to pass an include_headers list when precise control
  #: is needed.
  #:
  #: @param selector: the DKIM selector value for the signature
  #: @param domain: the DKIM domain value for the signature
  #: @param privkey: a PKCS#1 private key in base64-encoded text form
  #: @param identity: the DKIM identity value for the signature
  #: (default "@"+domain)
  #: @param canonicalize: the canonicalization algorithms to use
  #: (default (Simple, Simple))
  #: @param include_headers: a list of strings indicating which headers
  #: are to be signed (default rfc4871 recommended headers)
  #: @param length: true if the l= tag should be included to indicate
  #: body length signed (default False).
  #: @return: DKIM-Signature header field terminated by '\r\n'
  #: @raise DKIMException: when the message, include_headers, or key are badly
  #: formed.
  def sign(self, selector, domain, privkey, identity=None,
        canonicalize=(b'relaxed',b'simple'), include_headers=None, length=False):
    try:
        pk = parse_pem_private_key(privkey)
    except UnparsableKeyError as e:
        raise KeyFormatError(str(e))

    if identity is not None and not identity.endswith(domain):
        raise ParameterError("identity must end with domain")

    canon_policy = CanonicalizationPolicy.from_c_value(
        b'/'.join(canonicalize))
    headers = canon_policy.canonicalize_headers(self.headers)

    if include_headers is None:
        include_headers = self.default_sign_headers()

    # rfc4871 says FROM is required
    if b'from' not in ( x.lower() for x in include_headers ):
        raise ParameterError("The From header field MUST be signed")

    # raise exception for any SHOULD_NOT headers, call can modify 
    # SHOULD_NOT if really needed.
    for x in include_headers:
        if x.lower() in self.should_not_sign:
            raise ParameterError("The %s header field SHOULD NOT be signed"%x)

    body = canon_policy.canonicalize_body(self.body)

    hasher = HASH_ALGORITHMS[self.signature_algorithm]
    h = hasher()
    h.update(body)
    bodyhash = base64.b64encode(h.digest())

    sigfields = [x for x in [
        (b'v', b"1"),
        (b'a', self.signature_algorithm),
        (b'c', canon_policy.to_c_value()),
        (b'd', domain),
        (b'i', identity or b"@"+domain),
        length and (b'l', len(body)),
        (b'q', b"dns/txt"),
        (b's', selector),
        (b't', str(int(time.time())).encode('ascii')),
        (b'h', b" : ".join(include_headers)),
        (b'bh', bodyhash),
        # Force b= to fold onto it's own line so that refolding after
        # adding sig doesn't change whitespace for previous tags.
        (b'b', b'0'*60), 
    ] if x]
    include_headers = [x.lower() for x in include_headers]
    # record what verify should extract
    self.include_headers = tuple(include_headers)

    sig_value = fold(b"; ".join(b"=".join(x) for x in sigfields))
    sig_value = RE_BTAG.sub(b'\\1',sig_value)
    dkim_header = (b'DKIM-Signature', b' ' + sig_value)
    h = hasher()
    sig = dict(sigfields)
    self.signed_headers = hash_headers(
        h, canon_policy, headers, include_headers, dkim_header,sig)
    self.logger.debug("sign headers: %r" % self.signed_headers)

    try:
        sig2 = RSASSA_PKCS1_v1_5_sign(h, pk)
    except DigestTooLargeError:
        raise ParameterError("digest too large for modulus")
    # Folding b= is explicity allowed, but yahoo and live.com are broken
    #sig_value += base64.b64encode(bytes(sig2))
    # Instead of leaving unfolded (which lets an MTA fold it later and still
    # breaks yahoo and live.com), we change the default signing mode to
    # relaxed/simple (for broken receivers), and fold now.
    sig_value = fold(sig_value + base64.b64encode(bytes(sig2)))

    self.domain = domain
    self.selector = selector
    self.signature_fields = sig
    return b'DKIM-Signature: ' + sig_value + b"\r\n"

  #: Verify a DKIM signature.
  #: @type idx: int
  #: @param idx: which signature to verify.  The first (topmost) signature is 0.
  #: @type dnsfunc: callable
  #: @param dnsfunc: an option function to lookup TXT resource records
  #: for a DNS domain.  The default uses dnspython or pydns.
  #: @return: True if signature verifies or False otherwise
  #: @raise DKIMException: when the message, signature, or key are badly formed
  def verify(self,idx=0,dnsfunc=get_txt):

    sigheaders = [(x,y) for x,y in self.headers if x.lower() == b"dkim-signature"]
    if len(sigheaders) <= idx:
        return False

    # By default, we validate the first DKIM-Signature line found.
    try:
        sig = parse_tag_value(sigheaders[idx][1])
        self.signature_fields = sig
    except InvalidTagValueList as e:
        raise MessageFormatError(e)

    logger = self.logger
    logger.debug("sig: %r" % sig)

    validate_signature_fields(sig)
    self.domain = sig[b'd']
    self.selector = sig[b's']

    try:
        canon_policy = CanonicalizationPolicy.from_c_value(sig.get(b'c'))
    except InvalidCanonicalizationPolicyError as e:
        raise MessageFormatError("invalid c= value: %s" % e.args[0])
    headers = canon_policy.canonicalize_headers(self.headers)
    body = canon_policy.canonicalize_body(self.body)

    try:
        hasher = HASH_ALGORITHMS[sig[b'a']]
    except KeyError as e:
        raise MessageFormatError("unknown signature algorithm: %s" % e.args[0])

    if b'l' in sig:
        body = body[:int(sig[b'l'])]

    h = hasher()
    h.update(body)
    bodyhash = h.digest()
    logger.debug("bh: %s" % base64.b64encode(bodyhash))
    try:
        bh = base64.b64decode(re.sub(br"\s+", b"", sig[b'bh']))
    except TypeError as e:
        raise MessageFormatError(str(e))
    if bodyhash != bh:
        raise ValidationError(
            "body hash mismatch (got %s, expected %s)" %
            (base64.b64encode(bodyhash), sig[b'bh']))

    name = sig[b's'] + b"._domainkey." + sig[b'd'] + b"."
    s = dnsfunc(name)
    if not s:
        raise KeyFormatError("missing public key: %s"%name)
    try:
        pub = parse_tag_value(s)
    except InvalidTagValueList as e:
        raise KeyFormatError(e)
    try:
        pk = parse_public_key(base64.b64decode(pub[b'p']))
        self.keysize = bitsize(pk['modulus'])
    except KeyError:
        raise KeyFormatError("incomplete public key: %s" % s)
    except (TypeError,UnparsableKeyError) as e:
        raise KeyFormatError("could not parse public key (%s): %s" % (pub[b'p'],e))
    include_headers = [x.lower() for x in re.split(br"\s*:\s*", sig[b'h'])]
    self.include_headers = tuple(include_headers)
    # address bug#644046 by including any additional From header
    # fields when verifying.  Since there should be only one From header,
    # this shouldn't break any legitimate messages.  This could be
    # generalized to check for extras of other singleton headers.
    if b'from' in include_headers:
      include_headers.append(b'from')      
    h = hasher()
    self.signed_headers = hash_headers(
        h, canon_policy, headers, include_headers, sigheaders[idx], sig)
    try:
        signature = base64.b64decode(re.sub(br"\s+", b"", sig[b'b']))
        res = RSASSA_PKCS1_v1_5_verify(h, signature, pk)
        if res and self.keysize < self.minkey:
          raise KeyFormatError("public key too small: %d" % self.keysize)
        return res
    except (TypeError,DigestTooLargeError) as e:
        raise KeyFormatError("digest too large for modulus: %s"%e)

def sign(message, selector, domain, privkey, identity=None,
         canonicalize=(b'relaxed', b'simple'),
         signature_algorithm=b'rsa-sha256',
         include_headers=None, length=False, logger=None):
    """Sign an RFC822 message and return the DKIM-Signature header line.
    @param message: an RFC822 formatted message (with either \\n or \\r\\n line endings)
    @param selector: the DKIM selector value for the signature
    @param domain: the DKIM domain value for the signature
    @param privkey: a PKCS#1 private key in base64-encoded text form
    @param identity: the DKIM identity value for the signature (default "@"+domain)
    @param canonicalize: the canonicalization algorithms to use (default (Simple, Simple))
    @param include_headers: a list of strings indicating which headers are to be signed (default all headers not listed as SHOULD NOT sign)
    @param length: true if the l= tag should be included to indicate body length (default False)
    @param logger: a logger to which debug info will be written (default None)
    @return: DKIM-Signature header field terminated by \\r\\n
    @raise DKIMException: when the message, include_headers, or key are badly formed.
    """

    d = DKIM(message,logger=logger)
    if not include_headers:
        include_headers = d.all_sign_headers()
    return d.sign(selector, domain, privkey, identity=identity, canonicalize=canonicalize, include_headers=include_headers, length=length)

def verify(message, logger=None, dnsfunc=get_txt, minkey=1024):
    """Verify the first (topmost) DKIM signature on an RFC822 formatted message.
    @param message: an RFC822 formatted message (with either \\n or \\r\\n line endings)
    @param logger: a logger to which debug info will be written (default None)
    @return: True if signature verifies or False otherwise
    """
    d = DKIM(message,logger=logger,minkey=minkey)
    try:
        return d.verify(dnsfunc=dnsfunc)
    except DKIMException as x:
        if logger is not None:
            logger.error("%s" % x)
        return False



# -*- coding: utf-8 -*-
# adapted from https://github.com/kgn/cssutils/blob/master/examples/style.py
from __future__ import unicode_literals, print_function
import logging

from cssutils.css import CSSStyleSheet, CSSStyleDeclaration, CSSStyleRule
from cssutils import CSSParser
from lxml import etree

from emails.compat import to_unicode, string_types
import emails


# Workaround the missing python3-cssselect package
# If no system-installed cssselect library found, use one from our distribution
try:
    import cssselect
except ImportError:
    import sys, os.path
    my_packages = os.path.dirname(emails.packages.__file__)
    sys.path.insert(0, my_packages)
    import cssselect

from lxml.cssselect import CSSSelector, ExpressionError

class CSSInliner:

    NONVISUAL_TAGS = ['html', 'head', 'title', 'meta', 'link', 'script']

    DEBUG = False

    def __init__(self, base_url=None, css=None):

        self.stylesheet = CSSStyleSheet(href=base_url)
        self.base_url = base_url
        if css:
            self.add_css(css)

    def add_css(self, css, href=None):

        if isinstance(css, string_types):
            css = CSSParser().parseString(css, href=href)  # Распарсим файл

        for rule in css:
            self.stylesheet.add(rule)

    def log(self, level, *msg):
        if self.DEBUG:
            print(('%s- %s' % (level * '\t ', ' '.join((to_unicode(m or '') for m in msg)))))

    def styleattribute(self, element):
            """returns css.CSSStyleDeclaration of inline styles, for html: @style"""
            cssText = element.get('style')
            if cssText:
                try:
                    return CSSStyleDeclaration(cssText=cssText)
                except Exception as e:
                    # Sometimes here's error like "COLOR: ;"
                    logging.exception('Exception in styleattribute %s', cssText)
                    return None
            else:
                return None

    def getView(self, document, sheet, media='all', name=None, styleCallback=None):
        """
        document
            a DOM document, currently an lxml HTML document
        sheet
            a CSS StyleSheet object, currently cssutils sheet
        media: optional
            TODO: view for which media it should be
        name: optional
            TODO: names of sheets only
        styleCallback: optional
            should return css.CSSStyleDeclaration of inline styles, for html
            a style declaration for ``element@style``. Gets one parameter
            ``element`` which is the relevant DOMElement

        returns style view
            a dict of {DOMElement: css.CSSStyleDeclaration} for html
        """

        styleCallback = styleCallback or self.styleattribute

        _unmergable_rules = CSSStyleSheet()

        view = {}
        specificities = {}  # needed temporarily

        # TODO: filter rules simpler?, add @media
        rules = (rule for rule in sheet if rule.type == rule.STYLE_RULE)
        for rule in rules:
            for selector in rule.selectorList:
                self.log(0, 'SELECTOR', selector.selectorText)
                # TODO: make this a callback to be able to use other stuff than lxml
                try:
                    cssselector = CSSSelector(selector.selectorText)
                except (ExpressionError, NotImplementedError) as e:
                    _unmergable_rules.add(CSSStyleRule(selectorText=selector.selectorText,
                                                       style=rule.style))
                    continue

                matching = cssselector.evaluate(document)

                for element in matching:

                        if element.tag in self.NONVISUAL_TAGS:
                            continue

                        # add styles for all matching DOM elements
                        self.log(1, 'ELEMENT', id(element), element.text)

                        if element not in view:
                            # add initial empty style declatation
                            view[element] = CSSStyleDeclaration()
                            specificities[element] = {}

                            # and add inline @style if present
                            inlinestyle = styleCallback(element)
                            if inlinestyle:
                                for p in inlinestyle:
                                    # set inline style specificity
                                    view[element].setProperty(p)
                                    specificities[element][p.name] = (1, 0, 0, 0)

                        for p in rule.style:
                            # update style declaration
                            if p not in view[element]:
                                # setProperty needs a new Property object and
                                # MUST NOT reuse the existing Property
                                # which would be the same for all elements!
                                # see Issue #23
                                view[element].setProperty(p.name, p.value, p.priority)
                                specificities[element][p.name] = selector.specificity
                                self.log(2, view[element].getProperty('color'))

                            else:
                                self.log(2, view[element].getProperty('color'))
                                sameprio = (p.priority ==
                                            view[element].getPropertyPriority(p.name))
                                if not sameprio and bool(p.priority) or (
                                   sameprio and selector.specificity >=
                                        specificities[element][p.name]):
                                    # later, more specific or higher prio
                                    view[element].setProperty(p.name, p.value, p.priority)

        _unmergable_css = _unmergable_rules.cssText
        if _unmergable_css:
            e = etree.Element('style')
            e.text = to_unicode(_unmergable_css, 'utf-8')
            body = document.find('body') or document
            body.insert(0, e)  # add <style> right into body

        return view

    def transform(self, html):

        if isinstance(html, string_types):
            html = etree.HTML(html, parser=etree.HTMLParser())

        view = self.getView(html, self.stylesheet)

        # - add style into @style attribute
        for element, style in list(view.items()):
            v = style.getCssText(separator='')
            element.set('style', v)

        return html

    transform_html = transform  # compatibility



# encoding: utf-8
from __future__ import unicode_literals

__all__ = ['parse_name_and_email', 'load_email_charsets', 'MessageID']

import socket
import time
import os
import random
import email.charset

from email import generator
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.header import Header, decode_header
from email.utils import formataddr, parseaddr

from emails.compat import string_types, to_unicode, NativeStringIO, is_py2, BytesIO

_charsets_loaded = False

CHARSETS_FIX = [
    ['windows-1251', 'QP', 'QP'],
    # koi8 should by send as Quoted Printable because of bad SpamAssassin reaction on base64 (2008)
    ['koi8-r', 'QP', 'QP'],
    ['utf-8', 'BASE64', 'BASE64']
]


def load_email_charsets():
    global _charsets_loaded
    if not _charsets_loaded:
        for (charset, header_enc, body_enc) in CHARSETS_FIX:
            email.charset.add_charset(charset,
                                      getattr(email.charset, header_enc),
                                      getattr(email.charset, body_enc),
                                      charset)


# Django's CachedDnsName:
# Cached the hostname, but do it lazily: socket.getfqdn() can take a couple of
# seconds, which slows down the restart of the server.
class CachedDnsName(object):
    def __str__(self):
        return self.get_fqdn()

    def get_fqdn(self):
        if not hasattr(self, '_fqdn'):
            self._fqdn = socket.getfqdn()
        return self._fqdn


DNS_NAME = CachedDnsName()


def getheader(header_text, default="ascii"):
    """Decode the specified header"""
    headers = decode_header(header_text)
    header_sections = [unicode(text, charset or default)
                       for text, charset in headers]
    return u"".join(header_sections)


class MessageID:
    """Returns a string suitable for RFC 2822 compliant Message-ID, e.g:
    <20020201195627.33539.96671@nightshade.la.mastaler.com>
    Optional idstring if given is a string used to strengthen the
    uniqueness of the message id.
    Based on django.core.mail.message.make_msgid
    """

    def __init__(self, domain=None, idstring=None):
        self.domain = domain or DNS_NAME
        self.idstring = idstring

    def __call__(self):
        timeval = time.time()
        utcdate = time.strftime('%Y%m%d%H%M%S', time.gmtime(timeval))
        try:
            pid = os.getpid()
        except AttributeError:
            # No getpid() in Jython, for example.
            pid = 1
        randint = random.randrange(100000)
        if self.idstring is None:
            idstring = ''
        else:
            idstring = '.' + self.idstring
        msgid = '<%s.%s.%s%s@%s>' % (utcdate, pid, randint, idstring, self.domain)
        return msgid


def parse_name_and_email(obj, encoding='utf-8'):
    # In:  'john@smith.me' or  '"John Smith" <john@smith.me>' or ('John Smith', 'john@smith.me')
    # Out: (u'John Smith', u'john@smith.me')

    _realname = ''
    _email = ''

    if isinstance(obj, (list, tuple)):
        if len(obj) == 2:
            _realname, _email = obj

    elif isinstance(obj, string_types):
        _realname, _email = parseaddr(obj)

    else:
        raise ValueError("Can not parse_name_and_email from %s" % obj)

    if isinstance(_realname, bytes):
        _realname = to_unicode(_realname, encoding)

    if isinstance(_email, bytes):
        _email = to_unicode(_email, encoding)

    return _realname or None, _email or None


def sanitize_address(addr, encoding):
    if isinstance(addr, string_types):
        addr = parseaddr(to_unicode(addr))
    nm, addr = addr
    # This try-except clause is needed on Python 3 < 3.2.4
    # http://bugs.python.org/issue14291
    try:
        nm = Header(nm, encoding).encode()
    except UnicodeEncodeError:
        nm = Header(nm, 'utf-8').encode()
    try:
        addr.encode('ascii')
    except UnicodeEncodeError:  # IDN
        if '@' in addr:
            localpart, domain = addr.split('@', 1)
            localpart = str(Header(localpart, encoding))
            domain = domain.encode('idna').decode('ascii')
            addr = '@'.join([localpart, domain])
        else:
            addr = Header(addr, encoding).encode()
    return formataddr((nm, addr))


class MIMEMixin():
    def as_string(self, unixfrom=False, linesep='\n'):
        """Return the entire formatted message as a string.
        Optional `unixfrom' when True, means include the Unix From_ envelope
        header.
        This overrides the default as_string() implementation to not mangle
        lines that begin with 'From '. See bug #13433 for details.
        """
        fp = NativeStringIO()
        g = generator.Generator(fp, mangle_from_=False)
        if is_py2:
            g.flatten(self, unixfrom=unixfrom)
        else:
            g.flatten(self, unixfrom=unixfrom, linesep=linesep)

        return fp.getvalue()

    if is_py2:
        as_bytes = as_string
    else:
        def as_bytes(self, unixfrom=False, linesep='\n'):
            """Return the entire formatted message as bytes.
            Optional `unixfrom' when True, means include the Unix From_ envelope
            header.
            This overrides the default as_bytes() implementation to not mangle
            lines that begin with 'From '. See bug #13433 for details.
            """
            fp = BytesIO()
            g = generator.BytesGenerator(fp, mangle_from_=False)
            g.flatten(self, unixfrom=unixfrom, linesep=linesep)
            return fp.getvalue()


class SafeMIMEText(MIMEMixin, MIMEText):
    def __init__(self, text, subtype, charset):
        self.encoding = charset
        MIMEText.__init__(self, text, subtype, charset)

    def __setitem__(self, name, val):
        MIMEText.__setitem__(self, name, val)


class SafeMIMEMultipart(MIMEMixin, MIMEMultipart):

    def __init__(self, _subtype='mixed', boundary=None, _subparts=None, encoding=None, **_params):
        self.encoding = encoding
        MIMEMultipart.__init__(self, _subtype, boundary, _subparts, **_params)

    def __setitem__(self, name, val):
        MIMEMultipart.__setitem__(self, name, val)

def test_parse_name_and_email():
    assert parse_name_and_email('john@smith.me') == ('', 'john@smith.me')
    assert parse_name_and_email('"John Smith" <john@smith.me>') == \
           ('John Smith', 'john@smith.me')
    assert parse_name_and_email(['John Smith', 'john@smith.me']) == \
           ('John Smith', 'john@smith.me')



# encoding: utf-8

def simple_dict2str(d):
    # Simple dict serializer
    return ";".join( [ "%s=%s" % (k, v) for (k, v) in d.items() ] )

_serializer = simple_dict2str

class ObjectFactory:

    """
    Get object from cache or create new object.
    """

    def __init__(self, cls):
        self.cls = cls
        self.pool = {}

    def __getitem__(self, k):
        if not isinstance(k, dict):
            raise ValueError("item must be dict, not %s" % type(k))
        cache_key = _serializer(k)
        obj = self.pool.get(cache_key, None)
        if obj is None:
            obj = self.cls(**k)
            self.pool[cache_key] = obj
        return obj

    def invalidate(self, k):
        cache_key = _serializer(k)
        if cache_key in self.pool:
            del self.pool[cache_key]
        return self[cache_key]


# encoding: utf-8
from __future__ import unicode_literals
from .base import BaseTemplate


class MakoTemplate(BaseTemplate):

    def __init__(self, template_text, **kwargs):
        if 'mako_template' not in globals():
            globals()['mako_template'] = __import__('mako.template')
        self.template = mako_template.template.Template(template_text)

    def render(self, **kwargs):
        return self.template.render(**kwargs)


def test_mako_template_1():
    t = MakoTemplate("Hello, ${name}!")
    assert t.render(name='world') == "Hello, world!"


# encoding: utf-8
from __future__ import unicode_literals
from os.path import splitext
from emails.compat import OrderedDict, string_types

from .file import BaseFile


class FileStore(object):
    pass


class MemoryFileStore(FileStore):

    file_cls = BaseFile

    def __init__(self, file_cls=None):
        if file_cls:
            self.file_cls = file_cls
        self._files = OrderedDict()
        self._filenames = set()

    def __contains__(self, k):
        if isinstance(k, self.file_cls):
            return k.url in self._files
        elif isinstance(k, str):
            return k in self._files
        else:
            return False

    def keys(self):
        return list(self._files.keys())

    def __len__(self):
        return len(self._files)

    def as_dict(self):
        for d in self._files.values():
            yield d.as_dict()

    def remove(self, uri):
        if isinstance(uri, self.file_cls):
            uri = uri.uri

        assert isinstance(uri, string_types)

        v = self[uri]
        if v:
            filename = v.filename
            if filename and (filename in self._filenames):
                self._filenames.remove(filename)
            del self._files[uri]

    def unique_filename(self, filename):

        if filename not in self._filenames:
            return filename

        n = 1
        basefilename, ext = splitext(filename)

        while True:
            n += 1
            filename = "%s-%d%s" % (basefilename, n, ext)
            if filename not in self._filenames:
                return filename

    def add(self, value):

        if isinstance(value, self.file_cls):
            uri = value.uri
        elif isinstance(value, dict):
            value = self.file_cls(**value)
            uri = value.uri
        else:
            raise ValueError("Unknown file type: %s" % type(value))

        self.remove(uri)
        value.filename = self.unique_filename(value.filename)
        self._filenames.add(value.filename)
        self._files[uri] = value

    def by_uri(self, uri, synonims=None):
        r = self._files.get(uri, None)
        if r:
            return r
        if synonims:
            for _uri in synonims:
                r = self._files.get(_uri, None)
                if r:
                    return r
        return None

    def __getitem__(self, uri):
        return self._files.get(uri, None)

    def __iter__(self):
        for k in self._files:
            yield self._files[k]



# encoding: utf-8

__all__ = [ 'SMTPResponse', 'SMTPClientWithResponse', 'SMTPClientWithResponse_SSL' ]

from smtplib import _have_ssl, SMTP
import smtplib

import logging

logger = logging.getLogger(__name__)

class SMTPResponse(object):

    def __init__(self, host=None, port=None, ssl=None, exception=None):
        self.host = host
        self.port = port
        self.ssl = ssl
        self.responses = []
        self.exception = exception
        #self.complete = False
        self.success = None
        self.from_addr = None
        self.esmtp_opts = None
        self.rcpt_options = None
        self.to_addr = None
        self.status_code = None
        self.status_text = None
        self.last_command = None

    def set_status(self, command, code, text):
        self.responses.append( [command, code, text] )
        self.status_code = code
        self.status_text = text
        self.last_command = command

    def set_exception(self, exc):
        self.exception = exc

    def raise_if_needed():
        if self.exception:
            raise self.exception

    @property
    def error(self):
        return self.exception

    def __repr__(self):
        return "<emails.smtp.SMTPResponse status_code=%s status_text=%s>" % (self.status_code.__repr__(),
                                                                              self.status_text.__repr__())


#class SMTPCommandsLog:


class SMTPClientWithResponse(SMTP):

    def __init__(self, parent, **kwargs):
        self.parent = parent
        self.make_response = parent.make_response
        self._last_smtp_response = (None, None)
        self.tls = kwargs.pop('tls', False)
        self.debug = kwargs.pop('debug', False)
        self.user = kwargs.pop('user', None)
        self.password = kwargs.pop('password', None)
        SMTP.__init__(self, **kwargs)
        self.initialize()

    def initialize(self):
        if self.debug:
            self.set_debuglevel(1)
        if self.tls:
            self.ehlo()
            self.starttls()
            self.ehlo()
        if self.user:
            self.login(user=self.user, password=self.password)
        self.ehlo_or_helo_if_needed()
        return self

    def quit(self):
        """Closes the connection to the email server."""
        try:
            super(self, SMTPClientWithResponse).quit()
        except (smtplib.SMTPServerDisconnected, ):
            # This happens when calling quit() on a TLS connection
            # sometimes, or when the connection was already disconnected
            # by the server.
            self.close()

    def data(self, msg):
        (code, msg) = SMTP.data(self, msg)
        self._last_smtp_response = (code, msg)
        return code, msg


    def _send_one_mail(self, from_addr, to_addr, msg, mail_options=[], rcpt_options=[]):

        esmtp_opts = []
        if self.does_esmtp:
            if self.has_extn('size'):
                esmtp_opts.append("size=%d" % len(msg))
            for option in mail_options:
                esmtp_opts.append(option)

        response = self.make_response()

        response.from_addr = from_addr
        response.esmtp_opts = esmtp_opts[:]

        (code, resp) = self.mail(from_addr, esmtp_opts)
        response.set_status('mail', code, resp)

        if code != 250:
            self.rset()
            exc = smtplib.SMTPSenderRefused(code, resp, from_addr)
            response.set_exception(exc)
            return response

        response.to_addr = to_addr
        response.rcpt_options = rcpt_options[:]

        (code, resp) = self.rcpt(to_addr, rcpt_options)
        response.set_status('rcpt', code, resp)

        if (code != 250) and (code != 251):
            self.rset()
            exc = smtplib.SMTPRecipientsRefused(to_addr)
            response.set_exception(exc)
            return response

        (code, resp) = self.data(msg)
        response.set_status('data', code, resp)
        if code != 250:
            self.rset()
            exc = smtplib.SMTPDataError(code, resp)
            response.set_exception(exc)
            return response

        response.success = True
        return response

    def sendmail(self, from_addr, to_addrs, msg, mail_options=[], rcpt_options=[]):

        # Send one email and returns one response
        if not to_addrs:
            raise StopIteration

        assert isinstance(to_addrs, (list, tuple))

        if len(to_addrs)>1:
            logger.warning('Beware: emails.smtp.client.SMTPClientWithResponse.sendmail sends full message to each email')

        return [ self._send_one_mail(from_addr, to_addr, msg, mail_options, rcpt_options) \
                        for to_addr in to_addrs ]



if _have_ssl:

    from smtplib import SMTP_SSL
    import ssl

    class SMTPClientWithResponse_SSL(SMTPClientWithResponse, SMTP_SSL):

        def quit(self):
            """Closes the connection to the email server."""
            try:
                super(self, SMTPClientWithResponse_SSL).quit()
            except (ssl.SSLError, smtplib.SMTPServerDisconnected):
                # This happens when calling quit() on a TLS connection
                # sometimes, or when the connection was already disconnected
                # by the server.
                self.close()

else:

    class SMTPClientWithResponse_SSL:
        def __init__(self, *args, **kwargs):
            # should raise import error here
            import ssl






# encoding: utf-8

# This module use pydkim for DKIM signature

# Special note for python2.4 users:
#  - use dkimpy v0.3 from http://hewgill.com/pydkim/
#  - install hashlib (https://pypi.python.org/pypi/hashlib/20081119) and dnspython
from __future__ import unicode_literals
import logging

from emails.packages import dkim
from emails.compat import to_bytes, to_native


class DKIMSigner:

    def __init__(self, selector, domain, privkey, ignore_sign_errors=True, **kwargs):

        self.ignore_sign_errors = ignore_sign_errors
        self._sign_params = kwargs

        if privkey and hasattr(privkey, 'read'):
            privkey = privkey.read()

        self._sign_params.update({'privkey': to_bytes(privkey), 'domain': to_bytes(domain),
                                  'selector': to_bytes(selector)})

    def get_sign(self, message):

        dkim_header = None

        try:
            # pydkim module parses message and privkey on each signing
            # this is not optimal for mass operations
            # TODO: patch pydkim or use another signing module
            dkim_header = dkim.sign(message=message, **self._sign_params)
        except:
            if self.ignore_sign_errors:
                logging.exception('Error signing message')
            else:
                raise

        return to_native(dkim_header)

    def get_sign_header(self, message):

        dkim_header_str = self.get_sign(message)

        if dkim_header_str:
            # pdkim returns string, so we should split
            (header, value) = dkim_header_str.split(': ', 1)
            if value.endswith("\r\n"):
                value = value[:-2]
            return header, value

        return None



# encoding: utf-8
from __future__ import unicode_literals
__all__ = ['guess_charset', 'fix_content_type']

import re
import cgi
import chardet
from emails.compat import to_unicode

# HTML page charset stuff

RE_CHARSET = re.compile(b"charset=\"?'?(.+)\"?'?", re.I + re.S + re.M)
RE_META = re.compile(b"<meta.*?http-equiv=\"?'?content-type\"?'?.*?>", re.I + re.S + re.M)
RE_INSIDE_META = re.compile(b"content=\"?'?([^\"'>]+)", re.I + re.S + re.M)


def fix_content_type(content_type, t='image'):
    if not content_type:
        return "%s/unknown" % t
    else:
        return content_type


def guess_charset(headers, html):

    # guess by http headers
    if headers:
        #print(__name__, "guess_charset has headers", headers)
        content_type = headers['content-type']
        if content_type:
            _, params = cgi.parse_header(content_type)
            r = params.get('charset', None)
            if r:
                return r

    # guess by html meta
    #print(__name__, "guess_charset html=", html[:1024])
    for s in RE_META.findall(html):
        for x in RE_INSIDE_META.findall(s):
            for charset in RE_CHARSET.findall(x):
                return to_unicode(charset)

    # guess by chardet
    return chardet.detect(html)['encoding']


def set_content_type_meta(document, element_cls, content_type="text/html", charset="utf-8"):

    if document is None:
        document = element_cls('html')

    if document.tag!='html':
        html = element_cls('html')
        html.insert(0, document)
        document = html
    else:
        html = document

    head = document.find('head')
    if head is None:
        head = element_cls('head')
        html.insert(0, head)

    content_type_meta = None

    for meta in head.find('meta') or []:
        http_equiv = meta.get('http-equiv', None)
        if http_equiv and (http_equiv.lower() == 'content_type'):
            content_type_meta = meta
            break

    if content_type_meta is None:
        content_type_meta = element_cls('meta')
        head.append(content_type_meta)

    content_type_meta.set('content', '%s; charset=%s' % (content_type, charset))
    content_type_meta.set('http-equiv', "Content-Type")

    return document


def add_body_stylesheet(document, element_cls, cssText, tag="body"):

    style = element_cls('style')
    style.text = cssText

    body = document.find(tag)
    if body is None:
        body = document

    body.insert(0, style)

    return style



# encoding: utf-8
from __future__ import unicode_literals
import logging
import os
from os import path
import errno
from zipfile import ZipFile

from emails.compat import to_unicode, string_types

# FileSystemLoader adapted from jinja2.loaders


class FileNotFound(Exception):
    pass


def split_template_path(template):
    """Split a path into segments and perform a sanity check.  If it detects
    '..' in the path it will raise a `TemplateNotFound` error.
    """
    pieces = []
    for piece in template.split('/'):
        if path.sep in piece \
                or (path.altsep and path.altsep in piece) or \
                        piece == path.pardir:
            raise FileNotFound(template)
        elif piece and piece != '.':
            pieces.append(piece)
    return pieces


def open_if_exists(filename, mode='rb'):
    """Returns a file descriptor for the filename if that file exists,
    otherwise `None`.
    """
    try:
        return open(filename, mode)
    except IOError as e:
        if e.errno not in (errno.ENOENT, errno.EISDIR):
            raise


class BaseLoader(object):
    def __getitem__(self, filename):
        try:
            contents, _ = self.get_source(filename)
            return contents
        except FileNotFound:
            return None

    def find_index_file(self, filename=None):
        if filename:
            if self[filename]:
                return filename
            else:
                raise FileNotFound(filename)

        html_files = []

        for filename in self.list_files():

            f = path.basename(filename).lower()

            if f.endswith('.htm') or f.endswith('.html'):
                if f.startswith('index.'):
                    return filename
                else:
                    html_files.append(filename)

        if html_files:
            return html_files[0]

        raise FileNotFound('index html')


class FileSystemLoader(BaseLoader):
    """Loads templates from the file system.  This loader can find templates
    in folders on the file system and is the preferred way to load them.

    The loader takes the path to the templates as string, or if multiple
    locations are wanted a list of them which is then looked up in the
    given order:

    >>> loader = FileSystemLoader('/path/to/templates')
    >>> loader = FileSystemLoader(['/path/to/templates', '/other/path'])

    Per default the template encoding is ``'utf-8'`` which can be changed
    by setting the `encoding` parameter to something else.
    """

    def __init__(self, searchpath, encoding='utf-8', base_path=None):
        if isinstance(searchpath, string_types):
            searchpath = [searchpath]
        self.searchpath = list(searchpath)
        self.encoding = encoding
        self.base_path = base_path

    def get_source(self, template):

        if self.base_path:
            name = path.join(self.base_path, template)

        pieces = split_template_path(template)
        for searchpath in self.searchpath:
            filename = path.join(searchpath, *pieces)
            f = open_if_exists(filename)
            if f is None:
                continue
            try:
                contents = f.read().decode(self.encoding)
            finally:
                f.close()

            return contents, filename

        raise FileNotFound(template)

    def list_files(self):
        found = set()
        for searchpath in self.searchpath:
            for dirpath, dirnames, filenames in os.walk(searchpath):
                for filename in filenames:
                    template = path.join(dirpath, filename) \
                        [len(searchpath):].strip(path.sep) \
                        .replace(path.sep, '/')
                    if template[:2] == './':
                        template = template[2:]
                    if template not in found:
                        yield template


class ZipLoader(BaseLoader):
    def __init__(self, file, encoding='utf-8', base_path=None):
        self.zipfile = ZipFile(file, 'r')
        self.encoding = encoding
        self.base_path = base_path
        self.mapping = {}
        self._filenames = None

    def _decode_zip_filename(self, name):
        for enc in ('cp866', 'cp1251', 'utf-8'):
            try:
                return to_unicode(name, enc)
            except UnicodeDecodeError:
                pass
        return name

    def _unpack_zip(self):
        if self._filenames is None:
            self._filenames = {}
            for name in self.zipfile.namelist():
                decoded_name = self._decode_zip_filename(name)
                self._filenames[decoded_name] = name

    def get_source(self, name):

        logging.debug('ZipLoader.get_source %s', name)

        if self.base_path:
            name = path.join(self.base_path, name)
            logging.debug('ZipLoader.get_source has base_path, result name is %s', name)

        self._unpack_zip()

        if isinstance(name, str):
            name = to_unicode(name, 'utf-8')

        data = self.mapping.get(name, None)

        if data is not None:
            return data, name

        original_name = self._filenames.get(name)

        logging.debug('ZipLoader.get_source original_name=%s', original_name)

        if original_name is None:
            raise FileNotFound(name)

        data = self.zipfile.read(original_name)

        logging.debug('ZipLoader.get_source returns %s bytes', len(data))
        return data, name

    def list_files(self):
        self._unpack_zip()
        return sorted(self._filenames)



# encoding: utf-8
from __future__ import unicode_literals
from .base import BaseTemplate


class JinjaTemplate(BaseTemplate):

    DEFAULT_JINJA_ENVIRONMENT = {}

    def __init__(self, template_text, **kwargs):
        if 'jinja2' not in globals():
            globals()['jinja2'] = __import__('jinja2')
        self.environment = kwargs.get('environment', None) or jinja2.Environment(**self.DEFAULT_JINJA_ENVIRONMENT)
        self.template = self.environment.from_string(template_text)

    def render(self, **kwargs):
        return self.template.render(**kwargs)


def test_jinja_template_1():
    t = JinjaTemplate("Hello, {{name}}!")
    assert t.render(name='world') == "Hello, world!"
